@@ -66,7 +66,9 @@
 #include "distributed/multi_physical_planner.h"
 #include "distributed/multi_shard_transaction.h"
 #include "distributed/placement_connection.h"
+#include "distributed/relation_access_tracking.h"
 #include "distributed/remote_commands.h"
+#include "distributed/remote_transaction.h"
 #include "distributed/resource_lock.h"
 #include "distributed/shard_pruning.h"
 #include "distributed/version_compat.h"
@@ -799,12 +801,13 @@ OpenCopyConnections(CopyStmt *copyStatement, ShardConnections *shardConnections,
 	List *connectionList = NULL;
 	int64 shardId = shardConnections->shardId;
 	bool raiseInterrupts = true;
+	MemoryContext localContext =
+		AllocSetContextCreateExtended(CurrentMemoryContext,
+									  "OpenCopyConnections",
+									  ALLOCSET_DEFAULT_MINSIZE,
+									  ALLOCSET_DEFAULT_INITSIZE,
+									  ALLOCSET_DEFAULT_MAXSIZE);
 
-	MemoryContext localContext = AllocSetContextCreate(CurrentMemoryContext,
-													   "OpenCopyConnections",
-													   ALLOCSET_DEFAULT_MINSIZE,
-													   ALLOCSET_DEFAULT_INITSIZE,
-													   ALLOCSET_DEFAULT_MAXSIZE);
 
 	/* release finalized placement list at the end of this function */
 	MemoryContext oldContext = MemoryContextSwitchTo(localContext);
@@ -841,8 +844,9 @@ OpenCopyConnections(CopyStmt *copyStatement, ShardConnections *shardConnections,
 			}
 			else
 			{
-				ReportConnectionError(connection, WARNING);
-				MarkRemoteTransactionFailed(connection, true);
+				const bool raiseErrors = true;
+
+				HandleRemoteTransactionConnectionError(connection, raiseErrors);
 
 				failedPlacementCount++;
 				continue;
@@ -1233,9 +1237,11 @@ ReportCopyError(MultiConnection *connection, PGresult *result)
 	{
 		/* probably a constraint violation, show remote message and detail */
 		char *remoteDetail = PQresultErrorField(result, PG_DIAG_MESSAGE_DETAIL);
+		bool haveDetail = remoteDetail != NULL;
 
 		ereport(ERROR, (errmsg("%s", remoteMessage),
-						errdetail("%s", remoteDetail)));
+						haveDetail ? errdetail("%s", ApplyLogRedaction(remoteDetail)) :
+						0));
 	}
 	else
 	{
@@ -1245,7 +1251,7 @@ ReportCopyError(MultiConnection *connection, PGresult *result)
 		ereport(ERROR, (errcode(ERRCODE_IO_ERROR),
 						errmsg("failed to complete COPY on %s:%d", connection->hostname,
 							   connection->port),
-						errdetail("%s", remoteMessage)));
+						errdetail("%s", ApplyLogRedaction(remoteMessage))));
 	}
 }
 
@@ -1279,8 +1285,24 @@ ConversionPathForTypes(Oid inputType, Oid destType, CopyCoercionData *result)
 
 		case COERCION_PATH_ARRAYCOERCE:
 		{
-			ereport(ERROR, (errmsg("can not run query which uses an implicit coercion"
-								   " between array types")));
+			Oid inputBaseType = get_base_element_type(inputType);
+			Oid destBaseType = get_base_element_type(destType);
+			CoercionPathType baseCoercionType = COERCION_PATH_NONE;
+
+			if (inputBaseType != InvalidOid && destBaseType != InvalidOid)
+			{
+				baseCoercionType = find_coercion_pathway(inputBaseType, destBaseType,
+														 COERCION_EXPLICIT,
+														 &coercionFuncId);
+			}
+
+			if (baseCoercionType != COERCION_PATH_COERCEVIAIO)
+			{
+				ereport(ERROR, (errmsg("can not run query which uses an implicit coercion"
+									   " between array types")));
+			}
+
+			/* fallthrough */
 		}
 
 		case COERCION_PATH_COERCEVIAIO:
@@ -2147,6 +2169,7 @@ CitusCopyDestReceiverStartup(DestReceiver *dest, int operation,
 	copyOutState->fe_msgbuf = makeStringInfo();
 	copyOutState->rowcontext = GetPerTupleMemoryContext(copyDest->executorState);
 	copyDest->copyOutState = copyOutState;
+	copyDest->multiShardCopy = false;
 
 	/* prepare functions to call on received tuples */
 	{
@@ -2291,6 +2314,24 @@ CitusCopyDestReceiverReceive(TupleTableSlot *slot, DestReceiver *dest)
 											   &shardConnectionsFound);
 	if (!shardConnectionsFound)
 	{
+		/*
+		 * Keep track of multi shard accesses before opening connection
+		 * the second shard.
+		 */
+		if (!copyDest->multiShardCopy && hash_get_num_entries(shardConnectionHash) == 2)
+		{
+			Oid relationId = copyDest->distributedRelationId;
+
+			/* mark as multi shard to skip doing the same thing over and over */
+			copyDest->multiShardCopy = true;
+
+			/* error out of conflicting COPY */
+			CheckConflictingParallelCopyAccesses(relationId);
+
+			/* when we see multiple shard connections, we mark COPY as parallel modify */
+			RecordParallelModifyAccess(relationId);
+		}
+
 		/* open connections and initiate COPY on shard placements */
 		OpenCopyConnections(copyStatement, shardConnections, stopOnFailure,
 							copyOutState->binary);
