@@ -10,7 +10,10 @@
 
 #include "postgres.h"
 
+#ifdef HAVE_POLL_H
 #include <poll.h>
+#endif
+
 
 #include "libpq-fe.h"
 
@@ -19,46 +22,52 @@
 #include "access/hash.h"
 #include "commands/dbcommands.h"
 #include "distributed/connection_management.h"
+#include "distributed/errormessage.h"
 #include "distributed/metadata_cache.h"
 #include "distributed/hash_helpers.h"
 #include "distributed/placement_connection.h"
+#include "distributed/remote_commands.h"
+#include "distributed/version_compat.h"
 #include "mb/pg_wchar.h"
 #include "utils/hsearch.h"
 #include "utils/memutils.h"
 
 
 int NodeConnectionTimeout = 5000;
-int CitusSSLMode = CITUS_SSL_MODE_PREFER;
 HTAB *ConnectionHash = NULL;
+HTAB *ConnParamsHash = NULL;
 MemoryContext ConnectionContext = NULL;
 
-
 static uint32 ConnectionHashHash(const void *key, Size keysize);
 static int ConnectionHashCompare(const void *a, const void *b, Size keysize);
 static MultiConnection * StartConnectionEstablishment(ConnectionHashKey *key);
 static void AfterXactHostConnectionHandling(ConnectionHashEntry *entry, bool isCommit);
+static void DefaultCitusNoticeProcessor(void *arg, const char *message);
 static MultiConnection * FindAvailableConnection(dlist_head *connections, uint32 flags);
 
 
+static int CitusNoticeLogLevel = DEFAULT_CITUS_NOTICE_LEVEL;
+
+
 /*
  * Initialize per-backend connection management infrastructure.
  */
 void
 InitializeConnectionManagement(void)
 {
-	HASHCTL info;
+	HASHCTL info, connParamsInfo;
 	uint32 hashFlags = 0;
 
-
 	/*
 	 * Create a single context for connection and transaction related memory
 	 * management. Doing so, instead of allocating in TopMemoryContext, makes
 	 * it easier to associate used memory.
 	 */
-	ConnectionContext = AllocSetContextCreate(TopMemoryContext, "Connection Context",
-											  ALLOCSET_DEFAULT_MINSIZE,
-											  ALLOCSET_DEFAULT_INITSIZE,
-											  ALLOCSET_DEFAULT_MAXSIZE);
+	ConnectionContext = AllocSetContextCreateExtended(TopMemoryContext,
+													  "Connection Context",
+													  ALLOCSET_DEFAULT_MINSIZE,
+													  ALLOCSET_DEFAULT_INITSIZE,
+													  ALLOCSET_DEFAULT_MAXSIZE);
 
 	/* create (host,port,user,database) -> [connection] hash */
 	memset(&info, 0, sizeof(info));
@@ -69,8 +78,14 @@ InitializeConnectionManagement(void)
 	info.hcxt = ConnectionContext;
 	hashFlags = (HASH_ELEM | HASH_FUNCTION | HASH_CONTEXT | HASH_COMPARE);
 
+	memcpy(&connParamsInfo, &info, sizeof(HASHCTL));
+	connParamsInfo.entrysize = sizeof(ConnParamsHashEntry);
+
 	ConnectionHash = hash_create("citus connection cache (host,port,user,database)",
 								 64, &info, hashFlags);
+
+	ConnParamsHash = hash_create("citus connparams cache (host,port,user,database)",
+								 64, &connParamsInfo, hashFlags);
 }
 
 
@@ -113,6 +128,50 @@ GetNodeConnection(uint32 flags, const char *hostname, int32 port)
 }
 
 
+/*
+ * GetNonDataAccessConnection() establishes a connection to remote node, using
+ * default user and database. The returned connection is guaranteed to not have
+ * been used for any data access over any placements.
+ *
+ * See StartNonDataAccessConnection for details.
+ */
+MultiConnection *
+GetNonDataAccessConnection(const char *hostname, int32 port)
+{
+	MultiConnection *connection;
+
+	connection = StartNonDataAccessConnection(hostname, port);
+
+	FinishConnectionEstablishment(connection);
+
+	return connection;
+}
+
+
+/*
+ * StartNonDataAccessConnection() initiates a connection that is
+ * guaranteed to not have been used for any data access over any
+ * placements.
+ *
+ * The returned connection is started with the default user and database.
+ */
+MultiConnection *
+StartNonDataAccessConnection(const char *hostname, int32 port)
+{
+	uint32 flags = 0;
+	MultiConnection *connection = StartNodeConnection(flags, hostname, port);
+
+	if (ConnectionUsedForAnyPlacements(connection))
+	{
+		flags = FORCE_NEW_CONNECTION;
+
+		connection = StartNodeConnection(flags, hostname, port);
+	}
+
+	return connection;
+}
+
+
 /*
  * StartNodeConnection initiates a connection to remote node, using default
  * user and database.
@@ -449,9 +508,12 @@ FinishConnectionEstablishment(MultiConnection *connection)
 		/* Loop, to handle poll() being interrupted by signals (EINTR) */
 		while (true)
 		{
-			struct pollfd pollFileDescriptor;
 			int pollResult = 0;
 
+			/* we use poll(2) if available, otherwise select(2) */
+#ifdef HAVE_POLL
+			struct pollfd pollFileDescriptor;
+
 			pollFileDescriptor.fd = PQsocket(connection->pgConn);
 			if (pollmode == PGRES_POLLING_READING)
 			{
@@ -469,6 +531,31 @@ FinishConnectionEstablishment(MultiConnection *connection)
 			 * poll() after signal arrival.
 			 */
 			pollResult = poll(&pollFileDescriptor, 1, checkIntervalMS);
+#else /* !HAVE_POLL */
+			fd_set readFileDescriptorSet;
+			fd_set writeFileDescriptorSet;
+			fd_set exceptionFileDescriptorSet;
+			int selectTimeoutUS = checkIntervalMS * 1000;
+			struct timeval selectTimeout = { 0, selectTimeoutUS };
+			int selectFileDescriptor = PQsocket(connection->pgConn);
+
+			FD_ZERO(&readFileDescriptorSet);
+			FD_ZERO(&writeFileDescriptorSet);
+			FD_ZERO(&exceptionFileDescriptorSet);
+
+			if (pollmode == PGRES_POLLING_READING)
+			{
+				FD_SET(selectFileDescriptor, &readFileDescriptorSet);
+			}
+			else if (pollmode == PGRES_POLLING_WRITING)
+			{
+				FD_SET(selectFileDescriptor, &writeFileDescriptorSet);
+			}
+
+			pollResult = (select) (selectFileDescriptor + 1, &readFileDescriptorSet,
+								   &writeFileDescriptorSet, &exceptionFileDescriptorSet,
+								   &selectTimeout);
+#endif /* HAVE_POLL */
 
 			if (pollResult == 0)
 			{
@@ -502,20 +589,27 @@ FinishConnectionEstablishment(MultiConnection *connection)
 				 */
 				break;
 			}
-			else if (pollResult != EINTR)
-			{
-				/* Retrying, signal interrupted. So check. */
-				CHECK_FOR_INTERRUPTS();
-			}
 			else
 			{
-				/*
-				 * We ERROR here, instead of just returning a failed
-				 * connection, because this shouldn't happen, and indicates a
-				 * programming error somewhere, not a network etc. issue.
-				 */
-				ereport(ERROR, (errcode_for_socket_access(),
-								errmsg("poll() failed: %m")));
+				int errorCode = errno;
+#ifdef WIN32
+				errorCode = WSAGetLastError();
+#endif
+				if (errorCode == EINTR)
+				{
+					/* Retrying, signal interrupted. So check. */
+					CHECK_FOR_INTERRUPTS();
+				}
+				else
+				{
+					/*
+					 * We ERROR here, instead of just returning a failed
+					 * connection, because this shouldn't happen, and indicates a
+					 * programming error somewhere, not a network etc. issue.
+					 */
+					ereport(ERROR, (errcode_for_socket_access(),
+									errmsg("poll()/select() failed: %m")));
+				}
 			}
 		}
 	}
@@ -590,30 +684,31 @@ ConnectionHashCompare(const void *a, const void *b, Size keysize)
 static MultiConnection *
 StartConnectionEstablishment(ConnectionHashKey *key)
 {
-	char nodePortString[12];
-	const char *clientEncoding = GetDatabaseEncodingName();
+	bool found = false;
 	MultiConnection *connection = NULL;
-	const char *sslmode = CitusSSLModeString();
+	ConnParamsHashEntry *entry = NULL;
 
-	const char *keywords[] = {
-		"host", "port", "dbname", "user", "sslmode",
-		"client_encoding", "fallback_application_name",
-		NULL
-	};
-	const char *values[] = {
-		key->hostname, nodePortString, key->database, key->user, sslmode,
-		clientEncoding, "citus", NULL
-	};
+	/* search our cache for precomputed connection settings */
+	entry = hash_search(ConnParamsHash, key, HASH_ENTER, &found);
+	if (!found || !entry->isValid)
+	{
+		/* if they're not found, compute them from GUC, runtime, etc. */
+		GetConnParams(key, &entry->keywords, &entry->values, ConnectionContext);
+
+		entry->isValid = true;
+	}
 
 	connection = MemoryContextAllocZero(ConnectionContext, sizeof(MultiConnection));
-	sprintf(nodePortString, "%d", key->port);
 
 	strlcpy(connection->hostname, key->hostname, MAX_NODE_LENGTH);
 	connection->port = key->port;
 	strlcpy(connection->database, key->database, NAMEDATALEN);
 	strlcpy(connection->user, key->user, NAMEDATALEN);
 
-	connection->pgConn = PQconnectStartParams(keywords, values, false);
+
+	connection->pgConn = PQconnectStartParams((const char **) entry->keywords,
+											  (const char **) entry->values,
+											  false);
 	connection->connectionStart = GetCurrentTimestamp();
 
 	/*
@@ -623,56 +718,12 @@ StartConnectionEstablishment(ConnectionHashKey *key)
 	 */
 	PQsetnonblocking(connection->pgConn, true);
 
+	SetCitusNoticeProcessor(connection);
+
 	return connection;
 }
 
 
-/*
- * CitusSSLModeString returns the current value of citus.sslmode.
- */
-char *
-CitusSSLModeString(void)
-{
-	switch (CitusSSLMode)
-	{
-		case CITUS_SSL_MODE_DISABLE:
-		{
-			return "disable";
-		}
-
-		case CITUS_SSL_MODE_ALLOW:
-		{
-			return "allow";
-		}
-
-		case CITUS_SSL_MODE_PREFER:
-		{
-			return "prefer";
-		}
-
-		case CITUS_SSL_MODE_REQUIRE:
-		{
-			return "require";
-		}
-
-		case CITUS_SSL_MODE_VERIFY_CA:
-		{
-			return "verify-ca";
-		}
-
-		case CITUS_SSL_MODE_VERIFY_FULL:
-		{
-			return "verify-full";
-		}
-
-		default:
-		{
-			ereport(ERROR, (errmsg("unrecognized value for citus.sslmode")));
-		}
-	}
-}
-
-
 /*
  * Close all remote connections if necessary anymore (i.e. not session
  * lifetime), or if in a failed state.
@@ -726,3 +777,83 @@ AfterXactHostConnectionHandling(ConnectionHashEntry *entry, bool isCommit)
 		}
 	}
 }
+
+
+/*
+ * SetCitusNoticeProcessor sets the NoticeProcessor to DefaultCitusNoticeProcessor
+ */
+void
+SetCitusNoticeProcessor(MultiConnection *connection)
+{
+	PQsetNoticeProcessor(connection->pgConn, DefaultCitusNoticeProcessor,
+						 connection);
+}
+
+
+/*
+ * SetCitusNoticeLevel is used to set the notice level for distributed
+ * queries.
+ */
+void
+SetCitusNoticeLevel(int level)
+{
+	CitusNoticeLogLevel = level;
+}
+
+
+/*
+ * UnsetCitusNoticeLevel sets the CitusNoticeLogLevel back to
+ * its default value.
+ */
+void
+UnsetCitusNoticeLevel()
+{
+	CitusNoticeLogLevel = DEFAULT_CITUS_NOTICE_LEVEL;
+}
+
+
+/*
+ * DefaultCitusNoticeProcessor is used to redirect worker notices
+ * from logfile to console.
+ */
+static void
+DefaultCitusNoticeProcessor(void *arg, const char *message)
+{
+	MultiConnection *connection = (MultiConnection *) arg;
+	char *nodeName = connection->hostname;
+	uint32 nodePort = connection->port;
+	char *trimmedMessage = TrimLogLevel(message);
+	char *level = strtok((char *) message, ":");
+
+	ereport(CitusNoticeLogLevel,
+			(errmsg("%s", ApplyLogRedaction(trimmedMessage)),
+			 errdetail("%s from %s:%d", level, nodeName, nodePort)));
+}
+
+
+/*
+ * TrimLogLevel returns a copy of the string with the leading log level
+ * and spaces removed such as
+ *      From:
+ *          INFO:  "normal2_102070": scanned 0 of 0 pages...
+ *      To:
+ *          "normal2_102070": scanned 0 of 0 pages...
+ */
+char *
+TrimLogLevel(const char *message)
+{
+	char *chompedMessage = pchomp(message);
+	size_t n;
+
+	n = 0;
+	while (n < strlen(chompedMessage) && chompedMessage[n] != ':')
+	{
+		n++;
+	}
+
+	do {
+		n++;
+	} while (n < strlen(chompedMessage) && chompedMessage[n] == ' ');
+
+	return chompedMessage + n;
+}
