@@ -40,11 +40,14 @@
 #include "utils/memutils.h"
 
 
-/* controls the connection type for multi shard update/delete queries */
+/*
+ * Controls the connection type for multi shard modifications, DDLs
+ * TRUNCATE and real-time SELECT queries.
+ */
 int MultiShardConnectionType = PARALLEL_CONNECTION;
 
 
-/* ocal function forward declarations */
+/* local function forward declarations */
 static Relation StubRelation(TupleDesc tupleDescriptor);
 
 
@@ -297,3 +300,17 @@ ExecutePlanIntoDestReceiver(PlannedStmt *queryPlan, ParamListInfo params,
 #endif
 	PortalDrop(portal, false);
 }
+
+
+/*
+ * SetLocalMultiShardModifyModeToSequential simply a C interface for
+ * setting the following:
+ *      SET LOCAL citus.multi_shard_modify_mode = 'sequential';
+ */
+void
+SetLocalMultiShardModifyModeToSequential()
+{
+	set_config_option("citus.multi_shard_modify_mode", "sequential",
+					  (superuser() ? PGC_SUSET : PGC_USERSET), PGC_S_SESSION,
+					  GUC_ACTION_LOCAL, true, 0, false);
+}
