@@ -41,11 +41,13 @@
 #include "distributed/multi_router_planner.h"
 #include "distributed/multi_shard_transaction.h"
 #include "distributed/placement_connection.h"
+#include "distributed/relation_access_tracking.h"
 #include "distributed/subplan_execution.h"
 #include "distributed/relay_utility.h"
 #include "distributed/remote_commands.h"
 #include "distributed/remote_transaction.h"
 #include "distributed/resource_lock.h"
+#include "distributed/version_compat.h"
 #include "executor/execdesc.h"
 #include "executor/executor.h"
 #include "executor/instrument.h"
@@ -80,9 +82,12 @@ bool EnableDeadlockPrevention = true;
 static void AcquireMetadataLocks(List *taskList);
 static ShardPlacementAccess * CreatePlacementAccess(ShardPlacement *placement,
 													ShardPlacementAccessType accessType);
-static void ExecuteSingleModifyTask(CitusScanState *scanState, Task *task,
-									bool multipleTasks, bool expectResults);
+static int64 ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, CmdType
+									 operation, bool alwaysThrowErrorOnFailure, bool
+									 expectResults);
 static void ExecuteSingleSelectTask(CitusScanState *scanState, Task *task);
+static List * BuildPlacementAccessList(uint32 groupId, List *relationShardList,
+									   ShardPlacementAccessType accessType);
 static List * GetModifyConnections(Task *task, bool markCritical);
 static void ExecuteMultipleTasks(CitusScanState *scanState, List *taskList,
 								 bool isModificationQuery, bool expectResults);
@@ -91,16 +96,18 @@ static int64 ExecuteModifyTasks(List *taskList, bool expectResults,
 static void AcquireExecutorShardLock(Task *task, CmdType commandType);
 static void AcquireExecutorMultiShardLocks(List *taskList);
 static bool RequiresConsistentSnapshot(Task *task);
+static void RouterMultiModifyExecScan(CustomScanState *node);
+static void RouterSequentialModifyExecScan(CustomScanState *node);
 static void ExtractParametersFromParamListInfo(ParamListInfo paramListInfo,
 											   Oid **parameterTypes,
 											   const char ***parameterValues);
 static bool SendQueryInSingleRowMode(MultiConnection *connection, char *query,
 									 ParamListInfo paramListInfo);
-static bool StoreQueryResult(CitusScanState *scanState, MultiConnection *connection,
-							 bool failOnError, int64 *rows,
+static bool StoreQueryResult(CitusScanState *scanState, MultiConnection *connection, bool
+							 alwaysThrowErrorOnFailure, int64 *rows,
 							 DistributedExecutionStats *executionStats);
-static bool ConsumeQueryResult(MultiConnection *connection, bool failOnError,
-							   int64 *rows);
+static bool ConsumeQueryResult(MultiConnection *connection, bool
+							   alwaysThrowErrorOnFailure, int64 *rows);
 
 
 /*
@@ -149,7 +156,7 @@ AcquireExecutorShardLock(Task *task, CmdType commandType)
 	LOCKMODE lockMode = NoLock;
 	int64 shardId = task->anchorShardId;
 
-	if (commandType == CMD_SELECT || list_length(task->taskPlacementList) == 1)
+	if (commandType == CMD_SELECT)
 	{
 		/*
 		 * The executor shard lock is used to maintain consistency between
@@ -159,6 +166,28 @@ AcquireExecutorShardLock(Task *task, CmdType commandType)
 
 		lockMode = NoLock;
 	}
+	else if (list_length(task->taskPlacementList) == 1)
+	{
+		if (task->replicationModel == REPLICATION_MODEL_2PC)
+		{
+			/*
+			 * While we don't need a lock to ensure writes are applied in
+			 * a consistent order when there is a single replica. We also use
+			 * shard resource locks as a crude implementation of SELECT..
+			 * FOR UPDATE on reference tables, so we should always take
+			 * a lock that conflicts with the FOR UPDATE/SHARE locks.
+			 */
+			lockMode = RowExclusiveLock;
+		}
+		else
+		{
+			/*
+			 * When there is no replication, the worker itself can decide on
+			 * on the order in which writes are applied.
+			 */
+			lockMode = NoLock;
+		}
+	}
 	else if (AllModificationsCommutative)
 	{
 		/*
@@ -222,6 +251,55 @@ AcquireExecutorShardLock(Task *task, CmdType commandType)
 		LockShardResource(shardId, lockMode);
 	}
 
+	/*
+	 * If lock clause exists and it effects any reference table, we need to get
+	 * lock on shard resource. Type of lock is determined by the type of row lock
+	 * given in the query. If the type of row lock is either FOR NO KEY UPDATE or
+	 * FOR UPDATE we get ExclusiveLock on shard resource. We get ShareLock if it
+	 * is FOR KEY SHARE or FOR KEY SHARE.
+	 *
+	 * We have selected these lock types according to conflict table given in the
+	 * Postgres documentation. It is given that FOR UPDATE and FOR NO KEY UPDATE
+	 * must be conflict with each other modify command. By getting ExlcusiveLock
+	 * we guarantee that. Note that, getting ExlusiveLock does not mimic the
+	 * behaviour of Postgres exactly. Getting row lock with FOR NO KEY UPDATE and
+	 * FOR KEY SHARE do not conflicts in Postgres, yet they block each other in
+	 * our implementation. Since FOR SHARE and FOR KEY SHARE does not conflict
+	 * with each other but conflicts with modify commands, we get ShareLock for
+	 * them.
+	 */
+	if (task->relationRowLockList != NIL)
+	{
+		ListCell *rtiLockCell = NULL;
+		LOCKMODE rowLockMode = NoLock;
+
+		foreach(rtiLockCell, task->relationRowLockList)
+		{
+			RelationRowLock *relationRowLock = (RelationRowLock *) lfirst(rtiLockCell);
+			LockClauseStrength rowLockStrength = relationRowLock->rowLockStrength;
+			Oid relationId = relationRowLock->relationId;
+
+			if (PartitionMethod(relationId) == DISTRIBUTE_BY_NONE)
+			{
+				List *shardIntervalList = LoadShardIntervalList(relationId);
+				ShardInterval *referenceTableShardInterval = (ShardInterval *) linitial(
+					shardIntervalList);
+
+				if (rowLockStrength == LCS_FORKEYSHARE || rowLockStrength == LCS_FORSHARE)
+				{
+					rowLockMode = ShareLock;
+				}
+				else if (rowLockStrength == LCS_FORNOKEYUPDATE || rowLockStrength ==
+						 LCS_FORUPDATE)
+				{
+					rowLockMode = ExclusiveLock;
+				}
+
+				LockShardResource(referenceTableShardInterval->shardId, rowLockMode);
+			}
+		}
+	}
+
 	/*
 	 * If the task has a subselect, then we may need to lock the shards from which
 	 * the query selects as well to prevent the subselects from seeing different
@@ -309,8 +387,7 @@ AcquireExecutorMultiShardLocks(List *taskList)
 		/*
 		 * If the task has a subselect, then we may need to lock the shards from which
 		 * the query selects as well to prevent the subselects from seeing different
-		 * results on different replicas. In particular this prevents INSERT..SELECT
-		 * commands from having different effects on different placements.
+		 * results on different replicas.
 		 */
 
 		if (RequiresConsistentSnapshot(task))
@@ -329,18 +406,17 @@ AcquireExecutorMultiShardLocks(List *taskList)
 
 /*
  * RequiresConsistentSnapshot returns true if the given task need to take
- * the necessary locks to ensure that a subquery in the INSERT ... SELECT
- * query returns the same output for all task placements.
+ * the necessary locks to ensure that a subquery in the modify query
+ * returns the same output for all task placements.
  */
 static bool
 RequiresConsistentSnapshot(Task *task)
 {
 	bool requiresIsolation = false;
 
-	if (!task->insertSelectQuery)
+	if (!task->modifyWithSubquery)
 	{
 		/*
-		 * Only INSERT/SELECT commands currently require SELECT isolation.
 		 * Other commands do not read from other shards.
 		 */
 
@@ -439,6 +515,7 @@ CitusModifyBeginScan(CustomScanState *node, EState *estate, int eflags)
 			}
 
 			workerJob->taskList = taskList;
+			workerJob->partitionKeyValue = ExtractInsertPartitionKeyValue(jobQuery);
 		}
 
 		RebuildQueryStrings(jobQuery, taskList);
@@ -459,11 +536,15 @@ CitusModifyBeginScan(CustomScanState *node, EState *estate, int eflags)
 
 
 /*
- * RouterSequentialModifyExecScan executes 0 or more modifications on a
- * distributed table sequentially and returns results if there are any.
+ * RouterModifyExecScan executes a list of tasks on remote nodes, retrieves
+ * the results and, if RETURNING is used or SELECT FOR UPDATE executed,
+ * returns the results with a TupleTableSlot.
+ *
+ * The function can handle both single task query executions,
+ * sequential or parallel multi-task query executions.
  */
 TupleTableSlot *
-RouterSequentialModifyExecScan(CustomScanState *node)
+RouterModifyExecScan(CustomScanState *node)
 {
 	CitusScanState *scanState = (CitusScanState *) node;
 	TupleTableSlot *resultSlot = NULL;
@@ -471,28 +552,26 @@ RouterSequentialModifyExecScan(CustomScanState *node)
 	if (!scanState->finishedRemoteScan)
 	{
 		DistributedPlan *distributedPlan = scanState->distributedPlan;
-		bool hasReturning = distributedPlan->hasReturning;
 		Job *workerJob = distributedPlan->workerJob;
 		List *taskList = workerJob->taskList;
-		ListCell *taskCell = NULL;
-		bool multipleTasks = list_length(taskList) > 1;
+		bool parallelExecution = true;
 
-		/*
-		 * We could naturally handle function-based transactions (i.e. those using
-		 * PL/pgSQL or similar) by checking the type of queryDesc->dest, but some
-		 * customers already use functions that touch multiple shards from within
-		 * a function, so we'll ignore functions for now.
-		 */
-		if (IsTransactionBlock() || multipleTasks)
+		ExecuteSubPlans(distributedPlan);
+
+		if (list_length(taskList) <= 1 ||
+			IsMultiRowInsert(workerJob->jobQuery) ||
+			MultiShardConnectionType == SEQUENTIAL_CONNECTION)
 		{
-			BeginOrContinueCoordinatedTransaction();
+			parallelExecution = false;
 		}
 
-		foreach(taskCell, taskList)
+		if (parallelExecution)
 		{
-			Task *task = (Task *) lfirst(taskCell);
-
-			ExecuteSingleModifyTask(scanState, task, multipleTasks, hasReturning);
+			RouterMultiModifyExecScan(node);
+		}
+		else
+		{
+			RouterSequentialModifyExecScan(node);
 		}
 
 		scanState->finishedRemoteScan = true;
@@ -504,33 +583,141 @@ RouterSequentialModifyExecScan(CustomScanState *node)
 }
 
 
+/*
+ * RouterSequentialModifyExecScan executes 0 or more modifications on a
+ * distributed table sequentially and stores them in custom scan's tuple
+ * store. Note that we also use this path for SELECT ... FOR UPDATE queries.
+ */
+static void
+RouterSequentialModifyExecScan(CustomScanState *node)
+{
+	CitusScanState *scanState = (CitusScanState *) node;
+	DistributedPlan *distributedPlan = scanState->distributedPlan;
+	bool hasReturning = distributedPlan->hasReturning;
+	Job *workerJob = distributedPlan->workerJob;
+	List *taskList = workerJob->taskList;
+	ListCell *taskCell = NULL;
+	bool multipleTasks = list_length(taskList) > 1;
+	EState *executorState = scanState->customScanState.ss.ps.state;
+	bool taskListRequires2PC = TaskListRequires2PC(taskList);
+	bool alwaysThrowErrorOnFailure = false;
+	CmdType operation = scanState->distributedPlan->operation;
+
+	Assert(!scanState->finishedRemoteScan);
+
+	/*
+	 * We could naturally handle function-based transactions (i.e. those using
+	 * PL/pgSQL or similar) by checking the type of queryDesc->dest, but some
+	 * customers already use functions that touch multiple shards from within
+	 * a function, so we'll ignore functions for now.
+	 */
+	if (IsTransactionBlock() || multipleTasks || taskListRequires2PC)
+	{
+		BeginOrContinueCoordinatedTransaction();
+
+		/*
+		 * Although using two phase commit protocol is an independent decision than
+		 * failing on any error, we prefer to couple them. Our motivation is that
+		 * the failures are rare, and we prefer to avoid marking placements invalid
+		 * in case of failures.
+		 *
+		 * For reference tables, we always set alwaysThrowErrorOnFailure since we
+		 * absolutely want to avoid marking any placements invalid.
+		 *
+		 * We also cannot handle failures when there is RETURNING and there are more
+		 * than one task to execute.
+		 */
+		if (taskListRequires2PC)
+		{
+			CoordinatedTransactionUse2PC();
+
+			alwaysThrowErrorOnFailure = true;
+		}
+		else if (multipleTasks && hasReturning)
+		{
+			alwaysThrowErrorOnFailure = true;
+		}
+	}
+
+
+	foreach(taskCell, taskList)
+	{
+		Task *task = (Task *) lfirst(taskCell);
+		bool expectResults = (hasReturning || task->relationRowLockList != NIL);
+
+		executorState->es_processed +=
+			ExecuteSingleModifyTask(scanState, task, operation,
+									alwaysThrowErrorOnFailure, expectResults);
+	}
+}
+
+
+/*
+ * TaskListRequires2PC determines whether the given task list requires 2PC
+ * because the tasks provided operates on a reference table or there are multiple
+ * tasks and the commit protocol is 2PC.
+ *
+ * Note that we currently do not generate tasks lists that involves multiple different
+ * tables, thus we only check the first task in the list for reference tables.
+ */
+bool
+TaskListRequires2PC(List *taskList)
+{
+	Task *task = NULL;
+	bool multipleTasks = false;
+	uint64 anchorShardId = INVALID_SHARD_ID;
+
+	if (taskList == NIL)
+	{
+		return false;
+	}
+
+	task = (Task *) linitial(taskList);
+	if (task->replicationModel == REPLICATION_MODEL_2PC)
+	{
+		return true;
+	}
+
+	/*
+	 * Some tasks don't set replicationModel thus we rely on
+	 * the anchorShardId as well replicationModel.
+	 *
+	 * TODO: Do we ever need replicationModel in the Task structure?
+	 * Can't we always rely on anchorShardId?
+	 */
+	anchorShardId = task->anchorShardId;
+	if (ReferenceTableShardId(anchorShardId))
+	{
+		return true;
+	}
+
+	multipleTasks = list_length(taskList) > 1;
+	if (multipleTasks && MultiShardCommitProtocol == COMMIT_PROTOCOL_2PC)
+	{
+		return true;
+	}
+
+	return false;
+}
+
+
 /*
  * RouterMultiModifyExecScan executes a list of tasks on remote nodes, retrieves
  * the results and, if RETURNING is used, stores them in custom scan's tuple store.
- * Then, it returns tuples one by one from this tuple store.
  */
-TupleTableSlot *
+static void
 RouterMultiModifyExecScan(CustomScanState *node)
 {
 	CitusScanState *scanState = (CitusScanState *) node;
-	TupleTableSlot *resultSlot = NULL;
+	DistributedPlan *distributedPlan = scanState->distributedPlan;
+	Job *workerJob = distributedPlan->workerJob;
+	List *taskList = workerJob->taskList;
+	bool hasReturning = distributedPlan->hasReturning;
+	bool isModificationQuery = true;
 
-	if (!scanState->finishedRemoteScan)
-	{
-		DistributedPlan *distributedPlan = scanState->distributedPlan;
-		Job *workerJob = distributedPlan->workerJob;
-		List *taskList = workerJob->taskList;
-		bool hasReturning = distributedPlan->hasReturning;
-		bool isModificationQuery = true;
+	Assert(!scanState->finishedRemoteScan);
 
-		ExecuteMultipleTasks(scanState, taskList, isModificationQuery, hasReturning);
-
-		scanState->finishedRemoteScan = true;
-	}
-
-	resultSlot = ReturnTupleFromTuplestore(scanState);
-
-	return resultSlot;
+	ExecuteMultipleTasks(scanState, taskList, isModificationQuery, hasReturning);
 }
 
 
@@ -626,6 +813,13 @@ ExecuteSingleSelectTask(CitusScanState *scanState, Task *task)
 			placementAccessList = list_make1(placementAccess);
 		}
 
+		if (placementAccessList == NIL)
+		{
+			ereport(ERROR, (errcode(ERRCODE_T_R_SERIALIZATION_FAILURE),
+							errmsg("a placement was moved after the SELECT was "
+								   "planned")));
+		}
+
 		connection = GetPlacementListConnection(connectionFlags, placementAccessList,
 												NULL);
 
@@ -676,6 +870,28 @@ ExecuteSingleSelectTask(CitusScanState *scanState, Task *task)
  */
 List *
 BuildPlacementSelectList(uint32 groupId, List *relationShardList)
+{
+	return BuildPlacementAccessList(groupId, relationShardList, PLACEMENT_ACCESS_SELECT);
+}
+
+
+/*
+ * BuildPlacementDDLList is a warpper around BuildPlacementAccessList() for DDL access.
+ */
+List *
+BuildPlacementDDLList(uint32 groupId, List *relationShardList)
+{
+	return BuildPlacementAccessList(groupId, relationShardList, PLACEMENT_ACCESS_DDL);
+}
+
+
+/*
+ * BuildPlacementAccessList returns a list of placement accesses for the given
+ * relationShardList and the access type.
+ */
+static List *
+BuildPlacementAccessList(uint32 groupId, List *relationShardList,
+						 ShardPlacementAccessType accessType)
 {
 	ListCell *relationShardCell = NULL;
 	List *placementAccessList = NIL;
@@ -692,7 +908,7 @@ BuildPlacementSelectList(uint32 groupId, List *relationShardList)
 			continue;
 		}
 
-		placementAccess = CreatePlacementAccess(placement, PLACEMENT_ACCESS_SELECT);
+		placementAccess = CreatePlacementAccess(placement, accessType);
 		placementAccessList = lappend(placementAccessList, placementAccess);
 	}
 
@@ -719,20 +935,23 @@ CreatePlacementAccess(ShardPlacement *placement, ShardPlacementAccessType access
 
 /*
  * ExecuteSingleModifyTask executes the task on the remote node, retrieves the
- * results and stores them, if RETURNING is used, in a tuple store.
+ * results and stores them, if RETURNING is used, in a tuple store. The function
+ * can execute both DDL and DML tasks. When a DDL task is passed, the function
+ * does not expect scanState to be present.
  *
  * If the task fails on one of the placements, the function reraises the
  * remote error (constraint violation in DML), marks the affected placement as
  * invalid (other error on some placements, via the placement connection
  * framework), or errors out (failed on all placements).
+ *
+ * The function returns affectedTupleCount if applicable.
  */
-static void
-ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool multipleTasks,
-						bool expectResults)
+static int64
+ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, CmdType operation,
+						bool alwaysThrowErrorOnFailure, bool expectResults)
 {
-	CmdType operation = scanState->distributedPlan->operation;
-	EState *executorState = scanState->customScanState.ss.ps.state;
-	ParamListInfo paramListInfo = executorState->es_param_list_info;
+	EState *executorState = NULL;
+	ParamListInfo paramListInfo = NULL;
 	List *taskPlacementList = task->taskPlacementList;
 	List *connectionList = NIL;
 	ListCell *taskPlacementCell = NULL;
@@ -743,29 +962,23 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool multipleTask
 	bool gotResults = false;
 
 	char *queryString = task->queryString;
-	bool taskRequiresTwoPhaseCommit = (task->replicationModel == REPLICATION_MODEL_2PC);
 
 	ShardInterval *shardInterval = LoadShardInterval(task->anchorShardId);
 	Oid relationId = shardInterval->relationId;
 
-	/*
-	 * Modifications for reference tables are always done using 2PC. First
-	 * ensure that distributed transaction is started. Then force the
-	 * transaction manager to use 2PC while running the task on the
-	 * placements.
-	 */
-	if (taskRequiresTwoPhaseCommit)
+	if (scanState)
 	{
-		BeginOrContinueCoordinatedTransaction();
-		CoordinatedTransactionUse2PC();
+		executorState = scanState->customScanState.ss.ps.state;
+		paramListInfo = executorState->es_param_list_info;
 	}
 
 	/*
 	 * Get connections required to execute task. This will, if necessary,
 	 * establish the connection, mark as critical (when modifying reference
-	 * table) and start a transaction (when in a transaction).
+	 * table or multi-shard command) and start a transaction (when in a
+	 * transaction).
 	 */
-	connectionList = GetModifyConnections(task, taskRequiresTwoPhaseCommit);
+	connectionList = GetModifyConnections(task, alwaysThrowErrorOnFailure);
 
 	/*
 	 * If we are dealing with a partitioned table, we also need to lock its
@@ -779,8 +992,17 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool multipleTask
 		LockPartitionRelations(relationId, RowExclusiveLock);
 	}
 
-	/* prevent replicas of the same shard from diverging */
-	AcquireExecutorShardLock(task, operation);
+	/*
+	 * Prevent replicas of the same shard from diverging. We don't
+	 * need to acquire lock for TRUNCATE and DDLs since they already
+	 * acquire the necessary locks on the relations, and blocks any
+	 * unsafe concurrent operations.
+	 */
+	if (operation == CMD_INSERT || operation == CMD_UPDATE ||
+		operation == CMD_DELETE || operation == CMD_SELECT)
+	{
+		AcquireExecutorShardLock(task, operation);
+	}
 
 	/* try to execute modification on all placements */
 	forboth(taskPlacementCell, taskPlacementList, connectionCell, connectionList)
@@ -788,7 +1010,6 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool multipleTask
 		ShardPlacement *taskPlacement = (ShardPlacement *) lfirst(taskPlacementCell);
 		MultiConnection *connection = (MultiConnection *) lfirst(connectionCell);
 		bool queryOK = false;
-		bool failOnError = false;
 		int64 currentAffectedTupleCount = 0;
 
 		if (connection->remoteTransaction.transactionFailed)
@@ -811,25 +1032,13 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool multipleTask
 			continue;
 		}
 
-		/* if we're running a 2PC, the query should fail on error */
-		failOnError = taskRequiresTwoPhaseCommit;
-
-		if (multipleTasks && expectResults)
-		{
-			/*
-			 * If we have multiple tasks and one fails, we cannot clear
-			 * the tuple store and start over. Error out instead.
-			 */
-			failOnError = true;
-		}
-
 		if (failureCount + 1 == list_length(taskPlacementList))
 		{
 			/*
 			 * If we already failed on all other placements (possibly 0),
 			 * relay errors directly.
 			 */
-			failOnError = true;
+			alwaysThrowErrorOnFailure = true;
 		}
 
 		/*
@@ -839,12 +1048,12 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool multipleTask
 		 */
 		if (!gotResults && expectResults)
 		{
-			queryOK = StoreQueryResult(scanState, connection, failOnError,
+			queryOK = StoreQueryResult(scanState, connection, alwaysThrowErrorOnFailure,
 									   &currentAffectedTupleCount, NULL);
 		}
 		else
 		{
-			queryOK = ConsumeQueryResult(connection, failOnError,
+			queryOK = ConsumeQueryResult(connection, alwaysThrowErrorOnFailure,
 										 &currentAffectedTupleCount);
 		}
 
@@ -876,9 +1085,9 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool multipleTask
 
 	/*
 	 * If a command results in an error on all workers, we relay the last error
-	 * in the loop above by setting failOnError. However, if all connections fail
-	 * we still complete the loop without throwing an error. In that case, throw
-	 * an error below.
+	 * in the loop above by setting alwaysThrowErrorOnFailure. However, if all
+	 * connections fail we still complete the loop without throwing an error.
+	 * In that case, throw an error below.
 	 */
 	if (!resultsOK)
 	{
@@ -888,12 +1097,12 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool multipleTask
 	/* if some placements failed, ensure future statements don't access them */
 	MarkFailedShardPlacements();
 
-	executorState->es_processed += affectedTupleCount;
-
 	if (IsTransactionBlock())
 	{
 		XactModificationLevel = XACT_MODIFICATION_DATA;
 	}
+
+	return affectedTupleCount;
 }
 
 
@@ -916,20 +1125,44 @@ GetModifyConnections(Task *task, bool markCritical)
 	foreach(taskPlacementCell, taskPlacementList)
 	{
 		ShardPlacement *taskPlacement = (ShardPlacement *) lfirst(taskPlacementCell);
-		int connectionFlags = SESSION_LIFESPAN | FOR_DML;
+		int connectionFlags = SESSION_LIFESPAN;
 		MultiConnection *multiConnection = NULL;
 		List *placementAccessList = NIL;
 		ShardPlacementAccess *placementModification = NULL;
+		ShardPlacementAccessType accessType = PLACEMENT_ACCESS_DML;
+
+		if (task->taskType == DDL_TASK)
+		{
+			connectionFlags = connectionFlags | FOR_DDL;
+			accessType = PLACEMENT_ACCESS_DDL;
+		}
+		else
+		{
+			connectionFlags = connectionFlags | FOR_DML;
+			accessType = PLACEMENT_ACCESS_DML;
+		}
+
+		if (accessType == PLACEMENT_ACCESS_DDL)
+		{
+			/*
+			 * All relations appearing inter-shard DDL commands should be marked
+			 * with DDL access.
+			 */
+			placementAccessList =
+				BuildPlacementDDLList(taskPlacement->groupId, relationShardList);
+		}
+		else
+		{
+			/* create placement accesses for placements that appear in a subselect */
+			placementAccessList =
+				BuildPlacementSelectList(taskPlacement->groupId, relationShardList);
+		}
 
-		/* create placement accesses for placements that appear in a subselect */
-		placementAccessList = BuildPlacementSelectList(taskPlacement->groupId,
-													   relationShardList);
 
 		Assert(list_length(placementAccessList) == list_length(relationShardList));
 
 		/* create placement access for the placement that we're modifying */
-		placementModification = CreatePlacementAccess(taskPlacement,
-													  PLACEMENT_ACCESS_DML);
+		placementModification = CreatePlacementAccess(taskPlacement, accessType);
 		placementAccessList = lappend(placementAccessList, placementModification);
 
 		/* get an appropriate connection for the DML statement */
@@ -1010,23 +1243,57 @@ ExecuteModifyTasksWithoutResults(List *taskList)
 
 
 /*
- * ExecuteTasksSequentiallyWithoutResults basically calls ExecuteModifyTasks in
+ * ExecuteModifyTasksSequentiallyWithoutResults basically calls ExecuteSingleModifyTask in
  * a loop in order to simulate sequential execution of a list of tasks. Useful
  * in cases where issuing commands in parallel before waiting for results could
- * result in deadlocks (such as CREATE INDEX CONCURRENTLY).
+ * result in deadlocks (such as CREATE INDEX CONCURRENTLY or foreign key creation to
+ * reference tables).
+ *
+ * The function returns the affectedTupleCount if applicable. Otherwise, the function
+ * returns 0.
  */
-void
-ExecuteTasksSequentiallyWithoutResults(List *taskList)
+int64
+ExecuteModifyTasksSequentiallyWithoutResults(List *taskList, CmdType operation)
 {
 	ListCell *taskCell = NULL;
+	bool multipleTasks = list_length(taskList) > 1;
+	bool expectResults = false;
+	int64 affectedTupleCount = 0;
+	bool alwaysThrowErrorOnFailure = true;
+	bool taskListRequires2PC = TaskListRequires2PC(taskList);
 
+	/* decide on whether to use coordinated transaction and 2PC */
+	if (MultiShardCommitProtocol == COMMIT_PROTOCOL_BARE)
+	{
+		/* we don't run CREATE INDEX CONCURRENTLY in a distributed transaction */
+	}
+	else if (IsTransactionBlock() || multipleTasks)
+	{
+		BeginOrContinueCoordinatedTransaction();
+
+		if (taskListRequires2PC)
+		{
+			CoordinatedTransactionUse2PC();
+		}
+	}
+	else if (!multipleTasks && taskListRequires2PC)
+	{
+		/* DDL on a reference table should also use 2PC */
+		BeginOrContinueCoordinatedTransaction();
+		CoordinatedTransactionUse2PC();
+	}
+
+	/* now that we've decided on the transaction status, execute the tasks */
 	foreach(taskCell, taskList)
 	{
 		Task *task = (Task *) lfirst(taskCell);
-		List *singleTask = list_make1(task);
 
-		ExecuteModifyTasksWithoutResults(singleTask);
+		affectedTupleCount +=
+			ExecuteSingleModifyTask(NULL, task, operation, alwaysThrowErrorOnFailure,
+									expectResults);
 	}
+
+	return affectedTupleCount;
 }
 
 
@@ -1075,13 +1342,14 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 
 	/*
 	 * Ensure that there are no concurrent modifications on the same
-	 * shards. For DDL commands, we already obtained the appropriate
-	 * locks in ProcessUtility.
+	 * shards. In general, for DDL commands, we already obtained the
+	 * appropriate locks in ProcessUtility. However, we still prefer to
+	 * acquire the executor locks for DDLs specifically for TRUNCATE
+	 * command on a partition table since AcquireExecutorMultiShardLocks()
+	 * ensures that no concurrent modifications happens on the parent
+	 * tables.
 	 */
-	if (firstTask->taskType == MODIFY_TASK)
-	{
-		AcquireExecutorMultiShardLocks(taskList);
-	}
+	AcquireExecutorMultiShardLocks(taskList);
 
 	BeginOrContinueCoordinatedTransaction();
 
@@ -1091,7 +1359,33 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 		CoordinatedTransactionUse2PC();
 	}
 
-	if (firstTask->taskType == DDL_TASK)
+	/*
+	 * With a similar rationale as above, where we expect all tasks to operate on
+	 * the same relations, we prefer to record relation accesses for the first
+	 * task only.
+	 */
+	if (firstTask->taskType == MODIFY_TASK)
+	{
+		RecordRelationParallelModifyAccessForTask(firstTask);
+
+		/*
+		 * We prefer to mark with SELECT access as well because for multi shard
+		 * modification queries, the placement access list is always marked with both
+		 * DML and SELECT accesses.
+		 */
+		RecordRelationParallelSelectAccessForTask(firstTask);
+	}
+	else if (firstTask->taskType == DDL_TASK &&
+			 PartitionMethod(firstShardInterval->relationId) != DISTRIBUTE_BY_NONE)
+	{
+		/*
+		 * Even single task DDLs hit here, so we'd prefer
+		 * not to record for reference tables.
+		 */
+		RecordRelationParallelDDLAccessForTask(firstTask);
+	}
+
+	if (firstTask->taskType == DDL_TASK || firstTask->taskType == VACUUM_ANALYZE_TASK)
 	{
 		connectionFlags = FOR_DDL;
 	}
@@ -1153,7 +1447,7 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 			List *connectionList = NIL;
 			MultiConnection *connection = NULL;
 			int64 currentAffectedTupleCount = 0;
-			bool failOnError = true;
+			bool alwaysThrowErrorOnFailure = true;
 			bool queryOK PG_USED_FOR_ASSERTS_ONLY = false;
 
 			/* abort in case of cancellation */
@@ -1172,6 +1466,15 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 
 			connection = (MultiConnection *) list_nth(connectionList, placementIndex);
 
+			/*
+			 * if the task is a VACUUM or ANALYZE, we set CitusNoticeLogLevel to INFO
+			 * to see the logs in console.
+			 */
+			if (task->taskType == VACUUM_ANALYZE_TASK)
+			{
+				SetCitusNoticeLevel(INFO);
+			}
+
 			/*
 			 * If caller is interested, store query results the first time
 			 * through. The output of the query's execution on other shards is
@@ -1181,12 +1484,13 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 			{
 				Assert(scanState != NULL);
 
-				queryOK = StoreQueryResult(scanState, connection, failOnError,
+				queryOK = StoreQueryResult(scanState, connection,
+										   alwaysThrowErrorOnFailure,
 										   &currentAffectedTupleCount, NULL);
 			}
 			else
 			{
-				queryOK = ConsumeQueryResult(connection, failOnError,
+				queryOK = ConsumeQueryResult(connection, alwaysThrowErrorOnFailure,
 											 &currentAffectedTupleCount);
 			}
 
@@ -1234,6 +1538,9 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 		placementIndex++;
 	}
 
+	/* we should set the log level back to its default value since the task is done */
+	UnsetCitusNoticeLevel();
+
 	UnclaimAllShardConnections(shardConnectionHash);
 
 	CHECK_FOR_INTERRUPTS();
@@ -1260,6 +1567,9 @@ SendQueryInSingleRowMode(MultiConnection *connection, char *query,
 		Oid *parameterTypes = NULL;
 		const char **parameterValues = NULL;
 
+		/* force evaluation of bound params */
+		paramListInfo = copyParamList(paramListInfo);
+
 		ExtractParametersFromParamListInfo(paramListInfo, &parameterTypes,
 										   &parameterValues);
 
@@ -1273,16 +1583,18 @@ SendQueryInSingleRowMode(MultiConnection *connection, char *query,
 
 	if (querySent == 0)
 	{
-		MarkRemoteTransactionFailed(connection, false);
-		ReportConnectionError(connection, WARNING);
+		const bool raiseIfTransactionIsCritical = true;
+
+		HandleRemoteTransactionConnectionError(connection, raiseIfTransactionIsCritical);
 		return false;
 	}
 
 	singleRowMode = PQsetSingleRowMode(connection->pgConn);
 	if (singleRowMode == 0)
 	{
-		MarkRemoteTransactionFailed(connection, false);
-		ReportConnectionError(connection, WARNING);
+		const bool raiseIfTransactionIsCritical = true;
+
+		HandleRemoteTransactionConnectionError(connection, raiseIfTransactionIsCritical);
 		return false;
 	}
 
@@ -1368,7 +1680,7 @@ ExtractParametersFromParamListInfo(ParamListInfo paramListInfo, Oid **parameterT
  */
 static bool
 StoreQueryResult(CitusScanState *scanState, MultiConnection *connection,
-				 bool failOnError, int64 *rows,
+				 bool alwaysThrowErrorOnFailure, int64 *rows,
 				 DistributedExecutionStats *executionStats)
 {
 	TupleDesc tupleDescriptor =
@@ -1381,11 +1693,12 @@ StoreQueryResult(CitusScanState *scanState, MultiConnection *connection,
 	bool randomAccess = true;
 	bool interTransactions = false;
 	bool commandFailed = false;
-	MemoryContext ioContext = AllocSetContextCreate(CurrentMemoryContext,
-													"StoreQueryResult",
-													ALLOCSET_DEFAULT_MINSIZE,
-													ALLOCSET_DEFAULT_INITSIZE,
-													ALLOCSET_DEFAULT_MAXSIZE);
+	MemoryContext ioContext = AllocSetContextCreateExtended(CurrentMemoryContext,
+															"StoreQueryResult",
+															ALLOCSET_DEFAULT_MINSIZE,
+															ALLOCSET_DEFAULT_INITSIZE,
+															ALLOCSET_DEFAULT_MAXSIZE);
+
 	*rows = 0;
 
 	if (scanState->tuplestorestate == NULL)
@@ -1393,7 +1706,7 @@ StoreQueryResult(CitusScanState *scanState, MultiConnection *connection,
 		scanState->tuplestorestate =
 			tuplestore_begin_heap(randomAccess, interTransactions, work_mem);
 	}
-	else if (!failOnError)
+	else if (!alwaysThrowErrorOnFailure)
 	{
 		/* might have failed query execution on another placement before */
 		tuplestore_clear(scanState->tuplestorestate);
@@ -1423,6 +1736,10 @@ StoreQueryResult(CitusScanState *scanState, MultiConnection *connection,
 			int category = 0;
 			bool isConstraintViolation = false;
 
+			/*
+			 * Mark transaction as failed, but don't throw an error. This allows us
+			 * to give a more meaningful error message below.
+			 */
 			MarkRemoteTransactionFailed(connection, false);
 
 			/*
@@ -1433,7 +1750,8 @@ StoreQueryResult(CitusScanState *scanState, MultiConnection *connection,
 			category = ERRCODE_TO_CATEGORY(ERRCODE_INTEGRITY_CONSTRAINT_VIOLATION);
 			isConstraintViolation = SqlStateMatchesCategory(sqlStateString, category);
 
-			if (isConstraintViolation || failOnError)
+			if (isConstraintViolation || alwaysThrowErrorOnFailure ||
+				IsRemoteTransactionCritical(connection))
 			{
 				ReportResultError(connection, result, ERROR);
 			}
@@ -1515,7 +1833,8 @@ StoreQueryResult(CitusScanState *scanState, MultiConnection *connection,
  * has been an error.
  */
 static bool
-ConsumeQueryResult(MultiConnection *connection, bool failOnError, int64 *rows)
+ConsumeQueryResult(MultiConnection *connection, bool alwaysThrowErrorOnFailure,
+				   int64 *rows)
 {
 	bool commandFailed = false;
 	bool gotResponse = false;
@@ -1548,6 +1867,11 @@ ConsumeQueryResult(MultiConnection *connection, bool failOnError, int64 *rows)
 			int category = 0;
 			bool isConstraintViolation = false;
 
+			/*
+			 * Mark transaction as failed, but don't throw an error even if the
+			 * transaction is critical. This allows us to give a more meaningful
+			 * error message below.
+			 */
 			MarkRemoteTransactionFailed(connection, false);
 
 			/*
@@ -1558,7 +1882,8 @@ ConsumeQueryResult(MultiConnection *connection, bool failOnError, int64 *rows)
 			category = ERRCODE_TO_CATEGORY(ERRCODE_INTEGRITY_CONSTRAINT_VIOLATION);
 			isConstraintViolation = SqlStateMatchesCategory(sqlStateString, category);
 
-			if (isConstraintViolation || failOnError)
+			if (isConstraintViolation || alwaysThrowErrorOnFailure ||
+				IsRemoteTransactionCritical(connection))
 			{
 				ReportResultError(connection, result, ERROR);
 			}
