@@ -30,7 +30,10 @@
 #include "distributed/citus_ruleutils.h"
 #include "distributed/colocation_utils.h"
 #include "distributed/connection_management.h"
+#include "distributed/distributed_planner.h"
+#include "distributed/foreign_constraint.h"
 #include "distributed/multi_client_executor.h"
+#include "distributed/multi_router_executor.h"
 #include "distributed/master_metadata_utility.h"
 #include "distributed/master_protocol.h"
 #include "distributed/metadata_cache.h"
@@ -39,11 +42,13 @@
 #include "distributed/pg_dist_partition.h"
 #include "distributed/pg_dist_shard.h"
 #include "distributed/placement_connection.h"
+#include "distributed/relation_access_tracking.h"
 #include "distributed/remote_commands.h"
 #include "distributed/resource_lock.h"
 #include "distributed/transaction_management.h"
 #include "distributed/worker_manager.h"
 #include "distributed/worker_protocol.h"
+#include "distributed/version_compat.h"
 #include "storage/lmgr.h"
 #include "utils/builtins.h"
 #include "utils/fmgroids.h"
@@ -376,7 +381,6 @@ CreateAppendDistributedShardPlacements(Oid relationId, int64 shardId,
 	int placementsCreated = 0;
 	int attemptNumber = 0;
 	List *foreignConstraintCommandList = GetTableForeignConstraintCommands(relationId);
-	char *alterTableAttachPartitionCommand = NULL;
 	bool includeSequenceDefaults = false;
 	List *ddlCommandList = GetTableDDLEvents(relationId, includeSequenceDefaults);
 	uint32 connectionFlag = FOR_DDL;
@@ -411,8 +415,7 @@ CreateAppendDistributedShardPlacements(Oid relationId, int64 shardId,
 		}
 
 		WorkerCreateShard(relationId, shardIndex, shardId, ddlCommandList,
-						  foreignConstraintCommandList, alterTableAttachPartitionCommand,
-						  connection);
+						  foreignConstraintCommandList, connection);
 
 		InsertShardPlacementRow(shardId, INVALID_PLACEMENT_ID, shardState, shardSize,
 								nodeGroupId);
@@ -494,7 +497,7 @@ CreateShardsOnWorkers(Oid distributedRelationId, List *shardPlacements,
 	ListCell *connectionCell = NULL;
 	ListCell *shardPlacementCell = NULL;
 	int connectionFlags = FOR_DDL;
-	char *alterTableAttachPartitionCommand = NULL;
+	bool partitionTable = PartitionTable(distributedRelationId);
 
 	if (useExclusiveConnection)
 	{
@@ -502,12 +505,6 @@ CreateShardsOnWorkers(Oid distributedRelationId, List *shardPlacements,
 	}
 
 
-	if (PartitionTable(distributedRelationId))
-	{
-		alterTableAttachPartitionCommand =
-			GenerateAlterTableAttachPartitionCommand(distributedRelationId);
-	}
-
 	BeginOrContinueCoordinatedTransaction();
 
 	if (MultiShardCommitProtocol == COMMIT_PROTOCOL_2PC ||
@@ -516,6 +513,19 @@ CreateShardsOnWorkers(Oid distributedRelationId, List *shardPlacements,
 		CoordinatedTransactionUse2PC();
 	}
 
+	/* mark parallel relation accesses before opening connections */
+	if (ShouldRecordRelationAccess() && useExclusiveConnection)
+	{
+		RecordParallelDDLAccess(distributedRelationId);
+
+		/* we should mark the parent as well */
+		if (partitionTable)
+		{
+			Oid parentRelationId = PartitionParentOid(distributedRelationId);
+			RecordParallelDDLAccess(parentRelationId);
+		}
+	}
+
 	foreach(shardPlacementCell, shardPlacements)
 	{
 		ShardPlacement *shardPlacement = (ShardPlacement *) lfirst(shardPlacementCell);
@@ -529,8 +539,37 @@ CreateShardsOnWorkers(Oid distributedRelationId, List *shardPlacements,
 			shardIndex = ShardIndex(shardInterval);
 		}
 
-		connection = GetPlacementConnection(connectionFlags, shardPlacement,
-											placementOwner);
+		/*
+		 * For partitions, make sure that we mark the parent table relation access
+		 * with DDL. This is only important for parallel relation access in transaction
+		 * blocks, thus check useExclusiveConnection and transaction block as well.
+		 */
+		if (ShouldRecordRelationAccess() && useExclusiveConnection && partitionTable)
+		{
+			RelationShard *parentRelationShard = CitusMakeNode(RelationShard);
+			RelationShard *partitionRelationShard = CitusMakeNode(RelationShard);
+			List *relationShardList = NIL;
+			List *placementAccessList = NIL;
+
+			parentRelationShard->relationId = PartitionParentOid(distributedRelationId);
+			parentRelationShard->shardId =
+				ColocatedShardIdInRelation(parentRelationShard->relationId, shardIndex);
+			partitionRelationShard->relationId = distributedRelationId;
+			partitionRelationShard->shardId = shardId;
+
+			relationShardList = list_make2(parentRelationShard, partitionRelationShard);
+			placementAccessList = BuildPlacementDDLList(shardPlacement->groupId,
+														relationShardList);
+
+			connection = GetPlacementListConnection(connectionFlags, placementAccessList,
+													placementOwner);
+		}
+		else
+		{
+			connection = GetPlacementConnection(connectionFlags, shardPlacement,
+												placementOwner);
+		}
+
 		if (useExclusiveConnection)
 		{
 			ClaimConnectionExclusively(connection);
@@ -541,8 +580,7 @@ CreateShardsOnWorkers(Oid distributedRelationId, List *shardPlacements,
 		MarkRemoteTransactionCritical(connection);
 
 		WorkerCreateShard(distributedRelationId, shardIndex, shardId,
-						  ddlCommandList, foreignConstraintCommandList,
-						  alterTableAttachPartitionCommand, connection);
+						  ddlCommandList, foreignConstraintCommandList, connection);
 	}
 
 	/*
@@ -565,8 +603,7 @@ CreateShardsOnWorkers(Oid distributedRelationId, List *shardPlacements,
  */
 void
 WorkerCreateShard(Oid relationId, int shardIndex, uint64 shardId, List *ddlCommandList,
-				  List *foreignConstraintCommandList,
-				  char *alterTableAttachPartitionCommand, MultiConnection *connection)
+				  List *foreignConstraintCommandList, MultiConnection *connection)
 {
 	Oid schemaId = get_rel_namespace(relationId);
 	char *schemaName = get_namespace_name(schemaId);
@@ -625,12 +662,19 @@ WorkerCreateShard(Oid relationId, int shardIndex, uint64 shardId, List *ddlComma
 		 * In case of self referencing shards, relation itself might not be distributed
 		 * already. Therefore we cannot use ColocatedShardIdInRelation which assumes
 		 * given relation is distributed. Besides, since we know foreign key references
-		 * itself, referencedShardId is actual shardId anyway.
+		 * itself, referencedShardId is actual shardId anyway. Also, if the referenced
+		 * relation is a reference table, we cannot use ColocatedShardIdInRelation since
+		 * reference tables only have one shard. Instead, we fetch the one and only shard
+		 * from shardlist and use it.
 		 */
 		if (relationId == referencedRelationId)
 		{
 			referencedShardId = shardId;
 		}
+		else if (PartitionMethod(referencedRelationId) == DISTRIBUTE_BY_NONE)
+		{
+			referencedShardId = GetFirstShardId(referencedRelationId);
+		}
 		else
 		{
 			referencedShardId = ColocatedShardIdInRelation(referencedRelationId,
@@ -649,34 +693,11 @@ WorkerCreateShard(Oid relationId, int shardIndex, uint64 shardId, List *ddlComma
 	 * If the shard is created for a partition, send the command to create the
 	 * partitioning hierarcy on the shard.
 	 */
-	if (alterTableAttachPartitionCommand != NULL)
+	if (PartitionTable(relationId))
 	{
-		Oid parentRelationId = PartitionParentOid(relationId);
-		uint64 correspondingParentShardId = InvalidOid;
-		StringInfo applyAttachPartitionCommand = makeStringInfo();
-
-		Oid parentSchemaId = InvalidOid;
-		char *parentSchemaName = NULL;
-		char *escapedParentSchemaName = NULL;
-		char *escapedCommand = NULL;
-
-		Assert(PartitionTable(relationId));
-
-		parentSchemaId = get_rel_namespace(parentRelationId);
-		parentSchemaName = get_namespace_name(parentSchemaId);
-		escapedParentSchemaName = quote_literal_cstr(parentSchemaName);
-		escapedCommand = quote_literal_cstr(alterTableAttachPartitionCommand);
-
-		correspondingParentShardId = ColocatedShardIdInRelation(parentRelationId,
-																shardIndex);
-
-		appendStringInfo(applyAttachPartitionCommand,
-						 WORKER_APPLY_INTER_SHARD_DDL_COMMAND, correspondingParentShardId,
-						 escapedParentSchemaName, shardId, escapedSchemaName,
-						 escapedCommand);
-
-
-		ExecuteCriticalRemoteCommand(connection, applyAttachPartitionCommand->data);
+		ShardInterval *shardInterval = LoadShardInterval(shardId);
+		char *attachPartitionCommand = GenerateAttachShardPartitionCommand(shardInterval);
+		ExecuteCriticalRemoteCommand(connection, attachPartitionCommand);
 	}
 }
 
@@ -860,8 +881,8 @@ WorkerShardStats(ShardPlacement *placement, Oid relationId, char *shardName,
 
 	/* fill in the partition column name and shard name in the query. */
 	partitionColumn = PartitionColumn(relationId, unusedTableId);
-	partitionColumnName = get_attname(relationId, partitionColumn->varattno);
-
+	partitionColumnName = get_attname_internal(relationId, partitionColumn->varattno,
+											   false);
 	appendStringInfo(partitionValueQuery, SHARD_RANGE_QUERY,
 					 partitionColumnName, partitionColumnName, shardName);
 
