@@ -38,6 +38,7 @@
 #include "distributed/multi_server_executor.h"
 #include "distributed/listutils.h"
 #include "distributed/citus_ruleutils.h"
+#include "distributed/query_pushdown_planning.h"
 #include "distributed/relation_restriction_equivalence.h"
 #include "distributed/relay_utility.h"
 #include "distributed/resource_lock.h"
@@ -108,9 +109,13 @@ bool EnableRouterExecution = true;
 /* planner functions forward declarations */
 static DistributedPlan * CreateSingleTaskRouterPlan(Query *originalQuery,
 													Query *query,
-													RelationRestrictionContext *
-													restrictionContext);
+													PlannerRestrictionContext *
+													plannerRestrictionContext);
 static bool IsTidColumn(Node *node);
+static DeferredErrorMessage * MultiShardModifyQuerySupported(Query *originalQuery,
+															 PlannerRestrictionContext *
+															 plannerRestrictionContext);
+static bool HasDangerousJoinUsing(List *rtableList, Node *jtnode);
 static bool MasterIrreducibleExpression(Node *expression, bool *varArgument,
 										bool *badCoalesce);
 static bool MasterIrreducibleExpressionWalker(Node *expression, WalkerState *state);
@@ -124,13 +129,9 @@ static bool CanShardPrune(Oid distributedTableId, Query *query);
 static Job * CreateJob(Query *query);
 static Task * CreateTask(TaskType taskType);
 static Job * RouterJob(Query *originalQuery,
-					   RelationRestrictionContext *restrictionContext,
+					   PlannerRestrictionContext *plannerRestrictionContext,
 					   DeferredErrorMessage **planningError);
 static bool RelationPrunesToMultipleShards(List *relationShardList);
-static List * TargetShardIntervalsForRouter(Query *query,
-											RelationRestrictionContext *restrictionContext,
-											bool *multiShardQuery);
-static List * WorkersContainingAllShards(List *prunedShardIntervalsList);
 static void NormalizeMultiRowInsertTargetList(Query *query);
 static List * BuildRoutesForInsert(Query *query, DeferredErrorMessage **planningError);
 static List * GroupInsertValuesByShardId(List *insertValuesList);
@@ -138,9 +139,8 @@ static List * ExtractInsertValuesList(Query *query, Var *partitionColumn);
 static bool MultiRouterPlannableQuery(Query *query,
 									  RelationRestrictionContext *restrictionContext);
 static DeferredErrorMessage * ErrorIfQueryHasModifyingCTE(Query *queryTree);
-static RangeTblEntry * GetUpdateOrDeleteRTE(List *rangeTableList);
-static bool UpdateOrDeleteRTE(RangeTblEntry *rangeTableEntry);
-static bool SelectsFromDistributedTable(List *rangeTableList);
+static RangeTblEntry * GetUpdateOrDeleteRTE(Query *query);
+static bool SelectsFromDistributedTable(List *rangeTableList, Query *query);
 #if (PG_VERSION_NUM >= 100000)
 static List * get_all_actual_clauses(List *restrictinfo_list);
 #endif
@@ -149,10 +149,10 @@ static int CompareInsertValuesByShardId(const void *leftElement,
 static uint64 GetInitialShardId(List *relationShardList);
 static List * SingleShardSelectTaskList(Query *query, List *relationShardList,
 										List *placementList, uint64 shardId);
+static bool RowLocksOnRelations(Node *node, List **rtiLockList);
 static List * SingleShardModifyTaskList(Query *query, List *relationShardList,
 										List *placementList, uint64 shardId);
-static List * MultiShardModifyTaskList(Query *originalQuery, List *relationShardList,
-									   bool requiresMasterEvaluation);
+
 
 /*
  * CreateRouterPlan attempts to create a router executor plan for the given
@@ -161,12 +161,13 @@ static List * MultiShardModifyTaskList(Query *originalQuery, List *relationShard
  */
 DistributedPlan *
 CreateRouterPlan(Query *originalQuery, Query *query,
-				 RelationRestrictionContext *restrictionContext)
+				 PlannerRestrictionContext *plannerRestrictionContext)
 {
-	if (MultiRouterPlannableQuery(query, restrictionContext))
+	if (MultiRouterPlannableQuery(query,
+								  plannerRestrictionContext->relationRestrictionContext))
 	{
 		return CreateSingleTaskRouterPlan(originalQuery, query,
-										  restrictionContext);
+										  plannerRestrictionContext);
 	}
 
 	/*
@@ -193,7 +194,8 @@ CreateModifyPlan(Query *originalQuery, Query *query,
 	distributedPlan->operation = query->commandType;
 
 	distributedPlan->planningError = ModifyQuerySupported(query, originalQuery,
-														  multiShardQuery);
+														  multiShardQuery,
+														  plannerRestrictionContext);
 	if (distributedPlan->planningError != NULL)
 	{
 		return distributedPlan;
@@ -201,10 +203,7 @@ CreateModifyPlan(Query *originalQuery, Query *query,
 
 	if (UpdateOrDeleteQuery(query))
 	{
-		RelationRestrictionContext *restrictionContext =
-			plannerRestrictionContext->relationRestrictionContext;
-
-		job = RouterJob(originalQuery, restrictionContext,
+		job = RouterJob(originalQuery, plannerRestrictionContext,
 						&distributedPlan->planningError);
 	}
 	else
@@ -242,7 +241,7 @@ CreateModifyPlan(Query *originalQuery, Query *query,
  */
 static DistributedPlan *
 CreateSingleTaskRouterPlan(Query *originalQuery, Query *query,
-						   RelationRestrictionContext *restrictionContext)
+						   PlannerRestrictionContext *plannerRestrictionContext)
 {
 	Job *job = NULL;
 	DistributedPlan *distributedPlan = CitusMakeNode(DistributedPlan);
@@ -257,7 +256,8 @@ CreateSingleTaskRouterPlan(Query *originalQuery, Query *query,
 	}
 
 	/* we cannot have multi shard update/delete query via this code path */
-	job = RouterJob(originalQuery, restrictionContext, &distributedPlan->planningError);
+	job = RouterJob(originalQuery, plannerRestrictionContext,
+					&distributedPlan->planningError);
 
 	if (distributedPlan->planningError)
 	{
@@ -520,7 +520,8 @@ IsTidColumn(Node *node)
  * on the rewritten query.
  */
 DeferredErrorMessage *
-ModifyQuerySupported(Query *queryTree, Query *originalQuery, bool multiShardQuery)
+ModifyQuerySupported(Query *queryTree, Query *originalQuery, bool multiShardQuery,
+					 PlannerRestrictionContext *plannerRestrictionContext)
 {
 	Oid distributedTableId = ExtractFirstDistributedTableId(queryTree);
 	uint32 rangeTableId = 1;
@@ -529,12 +530,8 @@ ModifyQuerySupported(Query *queryTree, Query *originalQuery, bool multiShardQuer
 	List *rangeTableList = NIL;
 	ListCell *rangeTableCell = NULL;
 	uint32 queryTableCount = 0;
-	bool specifiesPartitionValue = false;
-	ListCell *setTargetCell = NULL;
-	List *onConflictSet = NIL;
-	Node *arbiterWhere = NULL;
-	Node *onConflictWhere = NULL;
 	CmdType commandType = queryTree->commandType;
+	DeferredErrorMessage *deferredError = NULL;
 
 	/*
 	 * Here, we check if a recursively planned query tries to modify
@@ -593,14 +590,21 @@ ModifyQuerySupported(Query *queryTree, Query *originalQuery, bool multiShardQuer
 
 		if (rangeTableEntry->rtekind == RTE_RELATION)
 		{
-			/*
-			 * We are sure that the table should be distributed, therefore no need to
-			 * call IsDistributedTable() here and DistributedTableCacheEntry will
-			 * error out if the table is not distributed
-			 */
-			DistTableCacheEntry *distTableEntry =
-				DistributedTableCacheEntry(rangeTableEntry->relid);
+			DistTableCacheEntry *distTableEntry = NULL;
 
+			if (!IsDistributedTable(rangeTableEntry->relid))
+			{
+				StringInfo errorMessage = makeStringInfo();
+				char *relationName = get_rel_name(rangeTableEntry->relid);
+
+				appendStringInfo(errorMessage, "relation %s is not distributed",
+								 relationName);
+
+				return DeferredError(ERRCODE_FEATURE_NOT_SUPPORTED,
+									 errorMessage->data, NULL, NULL);
+			}
+
+			distTableEntry = DistributedTableCacheEntry(rangeTableEntry->relid);
 			if (distTableEntry->partitionMethod == DISTRIBUTE_BY_NONE)
 			{
 				referenceTable = true;
@@ -691,23 +695,31 @@ ModifyQuerySupported(Query *queryTree, Query *originalQuery, bool multiShardQuer
 	}
 
 	/*
-	 * Reject queries which involve joins. Note that UPSERTs are exceptional for this case.
-	 * Queries like "INSERT INTO table_name ON CONFLICT DO UPDATE (col) SET other_col = ''"
-	 * contains two range table entries, and we have to allow them.
+	 * We have to allow modify queries with two range table entries, if it is pushdownable.
 	 */
-	if (commandType != CMD_INSERT && queryTableCount != 1)
+	if (commandType != CMD_INSERT)
 	{
-		/*
-		 * We support UPDATE and DELETE with joins unless they are multi shard
-		 * queries.
-		 */
-		if (!UpdateOrDeleteQuery(queryTree) || multiShardQuery)
+		/* We can not get restriction context via master_modify_multiple_shards path */
+		if (plannerRestrictionContext == NULL)
 		{
-			return DeferredError(ERRCODE_FEATURE_NOT_SUPPORTED,
-								 "cannot perform distributed planning for the given "
-								 "modification",
-								 "Joins are not supported in distributed "
-								 "modifications.", NULL);
+			if (queryTableCount != 1)
+			{
+				return DeferredError(ERRCODE_FEATURE_NOT_SUPPORTED,
+									 "cannot run multi shard modify query with master_modify_multiple_shards when the query involves subquery or join",
+									 "Execute the query without using master_modify_multiple_shards()",
+									 NULL);
+			}
+		}
+		/* If it is a multi-shard modify query with multiple tables */
+		else if (multiShardQuery)
+		{
+			DeferredErrorMessage *errorMessage = MultiShardModifyQuerySupported(
+				originalQuery, plannerRestrictionContext);
+
+			if (errorMessage != NULL)
+			{
+				return errorMessage;
+			}
 		}
 	}
 
@@ -753,7 +765,10 @@ ModifyQuerySupported(Query *queryTree, Query *originalQuery, bool multiShardQuer
 				TargetEntryChangesValue(targetEntry, partitionColumn,
 										queryTree->jointree))
 			{
-				specifiesPartitionValue = true;
+				return DeferredError(ERRCODE_FEATURE_NOT_SUPPORTED,
+									 "modifying the partition value of rows is not "
+									 "allowed",
+									 NULL, NULL);
 			}
 
 			if (commandType == CMD_UPDATE &&
@@ -813,13 +828,46 @@ ModifyQuerySupported(Query *queryTree, Query *originalQuery, bool multiShardQuer
 		}
 	}
 
-	if (commandType == CMD_INSERT && queryTree->onConflict != NULL)
+	deferredError = ErrorIfOnConflictNotSupported(queryTree);
+	if (deferredError != NULL)
 	{
-		onConflictSet = queryTree->onConflict->onConflictSet;
-		arbiterWhere = queryTree->onConflict->arbiterWhere;
-		onConflictWhere = queryTree->onConflict->onConflictWhere;
+		return deferredError;
 	}
 
+	return NULL;
+}
+
+
+/*
+ * ErrorIfOnConflictNotSupprted returns an error if an INSERT query has an
+ * unsupported ON CONFLICT clause. In particular, changing the partition
+ * column value or using volatile functions is not allowed.
+ */
+DeferredErrorMessage *
+ErrorIfOnConflictNotSupported(Query *queryTree)
+{
+	Oid distributedTableId = InvalidOid;
+	uint32 rangeTableId = 1;
+	Var *partitionColumn = NULL;
+	List *onConflictSet = NIL;
+	Node *arbiterWhere = NULL;
+	Node *onConflictWhere = NULL;
+	ListCell *setTargetCell = NULL;
+	bool specifiesPartitionValue = false;
+
+	CmdType commandType = queryTree->commandType;
+	if (commandType != CMD_INSERT || queryTree->onConflict == NULL)
+	{
+		return NULL;
+	}
+
+	distributedTableId = ExtractFirstDistributedTableId(queryTree);
+	partitionColumn = PartitionColumn(distributedTableId, rangeTableId);
+
+	onConflictSet = queryTree->onConflict->onConflictSet;
+	arbiterWhere = queryTree->onConflict->arbiterWhere;
+	onConflictWhere = queryTree->onConflict->onConflictWhere;
+
 	/*
 	 * onConflictSet is expanded via expand_targetlist() on the standard planner.
 	 * This ends up adding all the columns to the onConflictSet even if the user
@@ -902,6 +950,121 @@ ModifyQuerySupported(Query *queryTree, Query *originalQuery, bool multiShardQuer
 }
 
 
+/*
+ * MultiShardModifyQuerySupported returns the error message if the modify query is
+ * not pushdownable, otherwise it returns NULL.
+ */
+static DeferredErrorMessage *
+MultiShardModifyQuerySupported(Query *originalQuery,
+							   PlannerRestrictionContext *plannerRestrictionContext)
+{
+	DeferredErrorMessage *errorMessage = NULL;
+	RangeTblEntry *resultRangeTable = rt_fetch(originalQuery->resultRelation,
+											   originalQuery->rtable);
+	Oid resultRelationOid = resultRangeTable->relid;
+	char resultPartitionMethod = PartitionMethod(resultRelationOid);
+
+	if (HasDangerousJoinUsing(originalQuery->rtable, (Node *) originalQuery->jointree))
+	{
+		errorMessage = DeferredError(ERRCODE_FEATURE_NOT_SUPPORTED,
+									 "a join with USING causes an internal naming conflict, use "
+									 "ON instead",
+									 NULL, NULL);
+	}
+	else if (FindNodeCheck((Node *) originalQuery, CitusIsVolatileFunction))
+	{
+		errorMessage = DeferredError(ERRCODE_FEATURE_NOT_SUPPORTED,
+									 "functions used in UPDATE queries on distributed "
+									 "tables must not be VOLATILE",
+									 NULL, NULL);
+	}
+	else if (resultPartitionMethod == DISTRIBUTE_BY_NONE)
+	{
+		errorMessage = DeferredError(ERRCODE_FEATURE_NOT_SUPPORTED,
+									 "only reference tables may be queried when targeting "
+									 "a reference table with multi shard UPDATE/DELETE queries "
+									 "with multiple tables ",
+									 NULL, NULL);
+	}
+	else
+	{
+		errorMessage = DeferErrorIfUnsupportedSubqueryPushdown(originalQuery,
+															   plannerRestrictionContext);
+	}
+
+	return errorMessage;
+}
+
+
+/*
+ * HasDangerousJoinUsing search jointree for unnamed JOIN USING. Check the
+ * implementation of has_dangerous_join_using in ruleutils.
+ */
+static bool
+HasDangerousJoinUsing(List *rtableList, Node *joinTreeNode)
+{
+	if (IsA(joinTreeNode, RangeTblRef))
+	{
+		/* nothing to do here */
+	}
+	else if (IsA(joinTreeNode, FromExpr))
+	{
+		FromExpr *fromExpr = (FromExpr *) joinTreeNode;
+		ListCell *listCell;
+
+		foreach(listCell, fromExpr->fromlist)
+		{
+			if (HasDangerousJoinUsing(rtableList, (Node *) lfirst(listCell)))
+			{
+				return true;
+			}
+		}
+	}
+	else if (IsA(joinTreeNode, JoinExpr))
+	{
+		JoinExpr *joinExpr = (JoinExpr *) joinTreeNode;
+
+		/* Is it an unnamed JOIN with USING? */
+		if (joinExpr->alias == NULL && joinExpr->usingClause)
+		{
+			/*
+			 * Yes, so check each join alias var to see if any of them are not
+			 * simple references to underlying columns.  If so, we have a
+			 * dangerous situation and must pick unique aliases.
+			 */
+			RangeTblEntry *joinRTE = rt_fetch(joinExpr->rtindex, rtableList);
+			ListCell *listCell;
+
+			foreach(listCell, joinRTE->joinaliasvars)
+			{
+				Var *aliasVar = (Var *) lfirst(listCell);
+
+				if (aliasVar != NULL && !IsA(aliasVar, Var))
+				{
+					return true;
+				}
+			}
+		}
+
+		/* Nope, but inspect children */
+		if (HasDangerousJoinUsing(rtableList, joinExpr->larg))
+		{
+			return true;
+		}
+		if (HasDangerousJoinUsing(rtableList, joinExpr->rarg))
+		{
+			return true;
+		}
+	}
+	else
+	{
+		elog(ERROR, "unrecognized node type: %d",
+			 (int) nodeTag(joinTreeNode));
+	}
+	return false;
+}
+
+
 /*
  * UpdateOrDeleteQuery checks if the given query is an UPDATE or DELETE command.
  * If it is, it returns true otherwise it returns false.
@@ -1148,6 +1311,7 @@ RouterInsertJob(Query *originalQuery, Query *query, DeferredErrorMessage **plann
 	Job *job = NULL;
 	bool requiresMasterEvaluation = false;
 	bool deferredPruning = false;
+	Const *partitionKeyValue = NULL;
 
 	bool isMultiRowInsert = IsMultiRowInsert(query);
 	if (isMultiRowInsert)
@@ -1190,12 +1354,16 @@ RouterInsertJob(Query *originalQuery, Query *query, DeferredErrorMessage **plann
 	{
 		/* no functions or parameters, build the query strings upfront */
 		RebuildQueryStrings(originalQuery, taskList);
+
+		/* remember the partition column value */
+		partitionKeyValue = ExtractInsertPartitionKeyValue(originalQuery);
 	}
 
 	job = CreateJob(originalQuery);
 	job->taskList = taskList;
 	job->requiresMasterEvaluation = requiresMasterEvaluation;
 	job->deferredPruning = deferredPruning;
+	job->partitionKeyValue = partitionKeyValue;
 
 	return job;
 }
@@ -1351,12 +1519,12 @@ CreateTask(TaskType taskType)
 	task->upstreamTaskId = INVALID_TASK_ID;
 	task->shardInterval = NULL;
 	task->assignmentConstrained = false;
-	task->shardId = INVALID_SHARD_ID;
 	task->taskExecution = NULL;
 	task->upsertQuery = false;
 	task->replicationModel = REPLICATION_MODEL_INVALID;
+	task->relationRowLockList = NIL;
 
-	task->insertSelectQuery = false;
+	task->modifyWithSubquery = false;
 	task->relationShardList = NIL;
 
 	return task;
@@ -1398,18 +1566,18 @@ ExtractFirstDistributedTableId(Query *query)
  * multiple shard update/delete queries.
  */
 static Job *
-RouterJob(Query *originalQuery, RelationRestrictionContext *restrictionContext,
+RouterJob(Query *originalQuery, PlannerRestrictionContext *plannerRestrictionContext,
 		  DeferredErrorMessage **planningError)
 {
 	Job *job = NULL;
 	uint64 shardId = INVALID_SHARD_ID;
 	List *placementList = NIL;
 	List *relationShardList = NIL;
-	List *rangeTableList = NIL;
 	bool replacePrunedQueryWithDummy = false;
 	bool requiresMasterEvaluation = false;
 	RangeTblEntry *updateOrDeleteRTE = NULL;
 	bool isMultiShardModifyQuery = false;
+	Const *partitionKeyValue = NULL;
 
 	/* router planner should create task even if it deosn't hit a shard at all */
 	replacePrunedQueryWithDummy = true;
@@ -1417,19 +1585,20 @@ RouterJob(Query *originalQuery, RelationRestrictionContext *restrictionContext,
 	/* check if this query requires master evaluation */
 	requiresMasterEvaluation = RequiresMasterEvaluation(originalQuery);
 
-	(*planningError) = PlanRouterQuery(originalQuery, restrictionContext,
+	(*planningError) = PlanRouterQuery(originalQuery, plannerRestrictionContext,
 									   &placementList, &shardId, &relationShardList,
 									   replacePrunedQueryWithDummy,
-									   &isMultiShardModifyQuery);
+									   &isMultiShardModifyQuery,
+									   &partitionKeyValue);
 	if (*planningError)
 	{
 		return NULL;
 	}
 
 	job = CreateJob(originalQuery);
+	job->partitionKeyValue = partitionKeyValue;
 
-	ExtractRangeTableEntryWalker((Node *) originalQuery, &rangeTableList);
-	updateOrDeleteRTE = GetUpdateOrDeleteRTE(rangeTableList);
+	updateOrDeleteRTE = GetUpdateOrDeleteRTE(originalQuery);
 
 	/*
 	 * If all of the shards are pruned, we replace the relation RTE into
@@ -1451,7 +1620,10 @@ RouterJob(Query *originalQuery, RelationRestrictionContext *restrictionContext,
 	}
 	else if (isMultiShardModifyQuery)
 	{
-		job->taskList = MultiShardModifyTaskList(originalQuery, relationShardList,
+		job->taskList = QueryPushdownSqlTaskList(originalQuery, 0,
+												 plannerRestrictionContext->
+												 relationRestrictionContext,
+												 relationShardList, MODIFY_TASK,
 												 requiresMasterEvaluation);
 	}
 	else
@@ -1475,55 +1647,59 @@ SingleShardSelectTaskList(Query *query, List *relationShardList, List *placement
 {
 	Task *task = CreateTask(ROUTER_TASK);
 	StringInfo queryString = makeStringInfo();
+	List *relationRowLockList = NIL;
 
+	RowLocksOnRelations((Node *) query, &relationRowLockList);
 	pg_get_query_def(query, queryString);
 
 	task->queryString = queryString->data;
 	task->anchorShardId = shardId;
 	task->taskPlacementList = placementList;
 	task->relationShardList = relationShardList;
+	task->relationRowLockList = relationRowLockList;
 
 	return list_make1(task);
 }
 
 
 /*
- * MultiShardModifyTaskList generates task list for multi shard update/delete
- * queries.
+ * RowLocksOnRelations forms the list for range table IDs and corresponding
+ * row lock modes.
  */
-static List *
-MultiShardModifyTaskList(Query *originalQuery, List *relationShardList,
-						 bool requiresMasterEvaluation)
+static bool
+RowLocksOnRelations(Node *node, List **relationRowLockList)
 {
-	List *taskList = NIL;
-	ListCell *relationShardCell = NULL;
-	int taskId = 1;
-
-	foreach(relationShardCell, relationShardList)
+	if (node == NULL)
 	{
-		RelationShard *relationShard = (RelationShard *) lfirst(relationShardCell);
-		List *relationShardList = list_make1(relationShard);
-		Task *task = CreateTask(MODIFY_TASK);
-
-		if (!requiresMasterEvaluation)
-		{
-			Query *copiedQuery = copyObject(originalQuery);
-			StringInfo shardQueryString = makeStringInfo();
-
-			UpdateRelationToShardNames((Node *) copiedQuery, relationShardList);
-			pg_get_query_def(copiedQuery, shardQueryString);
-
-			task->queryString = shardQueryString->data;
-		}
-
-		task->taskId = taskId++;
-		task->anchorShardId = relationShard->shardId;
-		task->relationShardList = relationShardList;
-
-		taskList = lappend(taskList, task);
+		return false;
 	}
 
-	return taskList;
+	if (IsA(node, Query))
+	{
+		Query *query = (Query *) node;
+		ListCell *rowMarkCell = NULL;
+
+		foreach(rowMarkCell, query->rowMarks)
+		{
+			RowMarkClause *rowMarkClause = (RowMarkClause *) lfirst(rowMarkCell);
+			RangeTblEntry *rangeTable = rt_fetch(rowMarkClause->rti, query->rtable);
+			Oid relationId = rangeTable->relid;
+
+			if (IsDistributedTable(relationId))
+			{
+				RelationRowLock *relationRowLock = CitusMakeNode(RelationRowLock);
+				relationRowLock->relationId = relationId;
+				relationRowLock->rowLockStrength = rowMarkClause->strength;
+				*relationRowLockList = lappend(*relationRowLockList, relationRowLock);
+			}
+		}
+
+		return query_tree_walker(query, RowLocksOnRelations, relationRowLockList, 0);
+	}
+	else
+	{
+		return expression_tree_walker(node, RowLocksOnRelations, relationRowLockList);
+	}
 }
 
 
@@ -1543,13 +1719,13 @@ SingleShardModifyTaskList(Query *query, List *relationShardList, List *placement
 	RangeTblEntry *updateOrDeleteRTE = NULL;
 
 	ExtractRangeTableEntryWalker((Node *) query, &rangeTableList);
-	updateOrDeleteRTE = GetUpdateOrDeleteRTE(rangeTableList);
+	updateOrDeleteRTE = GetUpdateOrDeleteRTE(query);
 
 	modificationTableCacheEntry = DistributedTableCacheEntry(updateOrDeleteRTE->relid);
 	modificationPartitionMethod = modificationTableCacheEntry->partitionMethod;
 
 	if (modificationPartitionMethod == DISTRIBUTE_BY_NONE &&
-		SelectsFromDistributedTable(rangeTableList))
+		SelectsFromDistributedTable(rangeTableList, query))
 	{
 		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 						errmsg("cannot perform select on a distributed table "
@@ -1569,55 +1745,37 @@ SingleShardModifyTaskList(Query *query, List *relationShardList, List *placement
 
 
 /*
- * GetUpdateOrDeleteRTE walks over the given range table list, and checks if
- * it has an UPDATE or DELETE RTE. If it finds one, it return it immediately.
+ * GetUpdateOrDeleteRTE checks query if it has an UPDATE or DELETE RTE. If it finds
+ * it returns it.
  */
 static RangeTblEntry *
-GetUpdateOrDeleteRTE(List *rangeTableList)
+GetUpdateOrDeleteRTE(Query *query)
 {
-	ListCell *rangeTableCell = NULL;
-
-	foreach(rangeTableCell, rangeTableList)
+	if (query->resultRelation > 0)
 	{
-		RangeTblEntry *rangeTableEntry = (RangeTblEntry *) lfirst(rangeTableCell);
-
-		if (UpdateOrDeleteRTE(rangeTableEntry))
-		{
-			return rangeTableEntry;
-		}
+		return rt_fetch(query->resultRelation, query->rtable);
 	}
 
 	return NULL;
 }
 
 
-/*
- * UpdateOrDeleteRTE checks if the given range table entry is an UPDATE or
- * DELETE RTE by checking required permissions on it.
- */
-static bool
-UpdateOrDeleteRTE(RangeTblEntry *rangeTableEntry)
-{
-	if ((ACL_UPDATE & rangeTableEntry->requiredPerms) ||
-		(ACL_DELETE & rangeTableEntry->requiredPerms))
-	{
-		return true;
-	}
-	else
-	{
-		return false;
-	}
-}
-
-
 /*
  * SelectsFromDistributedTable checks if there is a select on a distributed
  * table by looking into range table entries.
  */
 static bool
-SelectsFromDistributedTable(List *rangeTableList)
+SelectsFromDistributedTable(List *rangeTableList, Query *query)
 {
 	ListCell *rangeTableCell = NULL;
+	int resultRelation = query->resultRelation;
+	RangeTblEntry *resultRangeTableEntry = NULL;
+
+	if (resultRelation > 0)
+	{
+		resultRangeTableEntry = rt_fetch(resultRelation, query->rtable);
+	}
+
 	foreach(rangeTableCell, rangeTableList)
 	{
 		RangeTblEntry *rangeTableEntry = (RangeTblEntry *) lfirst(rangeTableCell);
@@ -1630,7 +1788,8 @@ SelectsFromDistributedTable(List *rangeTableList)
 
 		cacheEntry = DistributedTableCacheEntry(rangeTableEntry->relid);
 		if (cacheEntry->partitionMethod != DISTRIBUTE_BY_NONE &&
-			!UpdateOrDeleteRTE(rangeTableEntry))
+			(resultRangeTableEntry == NULL || resultRangeTableEntry->relid !=
+			 rangeTableEntry->relid))
 		{
 			return true;
 		}
@@ -1662,9 +1821,11 @@ SelectsFromDistributedTable(List *rangeTableList)
  * 0 values in UpdateRelationToShardNames.
  */
 DeferredErrorMessage *
-PlanRouterQuery(Query *originalQuery, RelationRestrictionContext *restrictionContext,
+PlanRouterQuery(Query *originalQuery,
+				PlannerRestrictionContext *plannerRestrictionContext,
 				List **placementList, uint64 *anchorShardId, List **relationShardList,
-				bool replacePrunedQueryWithDummy, bool *multiShardModifyQuery)
+				bool replacePrunedQueryWithDummy, bool *multiShardModifyQuery,
+				Const **partitionValueConst)
 {
 	static uint32 zeroShardQueryRoundRobin = 0;
 
@@ -1676,12 +1837,14 @@ PlanRouterQuery(Query *originalQuery, RelationRestrictionContext *restrictionCon
 	bool shardsPresent = false;
 	uint64 shardId = INVALID_SHARD_ID;
 	CmdType commandType = originalQuery->commandType;
-	bool isMultiShardModifyQuery = false;
 
 	*placementList = NIL;
-	prunedRelationShardList = TargetShardIntervalsForRouter(originalQuery,
-															restrictionContext,
-															&isMultiShardQuery);
+
+	prunedRelationShardList = TargetShardIntervalsForQuery(originalQuery,
+														   plannerRestrictionContext->
+														   relationRestrictionContext,
+														   &isMultiShardQuery,
+														   partitionValueConst);
 
 	if (isMultiShardQuery)
 	{
@@ -1700,13 +1863,23 @@ PlanRouterQuery(Query *originalQuery, RelationRestrictionContext *restrictionCon
 		Assert(UpdateOrDeleteQuery(originalQuery));
 
 		planningError = ModifyQuerySupported(originalQuery, originalQuery,
-											 isMultiShardQuery);
+											 isMultiShardQuery,
+											 plannerRestrictionContext);
 		if (planningError != NULL)
 		{
 			return planningError;
 		}
-
-		isMultiShardModifyQuery = true;
+		else
+		{
+			/*
+			 * If the modify query uses multiple shards and update/delete query, relation
+			 * shard list should be returned as list of shard list for each table. Check
+			 * the implementation of QueryPushdownSqlTaskList.
+			 */
+			*relationShardList = prunedRelationShardList;
+			*multiShardModifyQuery = true;
+			return planningError;
+		}
 	}
 
 	foreach(prunedRelationShardListCell, prunedRelationShardList)
@@ -1734,12 +1907,6 @@ PlanRouterQuery(Query *originalQuery, RelationRestrictionContext *restrictionCon
 		}
 	}
 
-	if (isMultiShardModifyQuery)
-	{
-		*multiShardModifyQuery = true;
-		return planningError;
-	}
-
 	/*
 	 * We bail out if there are RTEs that prune multiple shards above, but
 	 * there can also be multiple RTEs that reference the same relation.
@@ -1852,7 +2019,7 @@ GetInitialShardId(List *relationShardList)
 
 
 /*
- * TargetShardIntervalsForRouter performs shard pruning for all referenced relations
+ * TargetShardIntervalsForQuery performs shard pruning for all referenced relations
  * in the query and returns list of shards per relation. Shard pruning is done based
  * on provided restriction context per relation. The function sets multiShardQuery
  * to true if any of the relations pruned down to more than one active shard. It
@@ -1861,13 +2028,15 @@ GetInitialShardId(List *relationShardList)
  * 'and 1=0', such queries are treated as if all of the shards of joining
  * relations are pruned out.
  */
-static List *
-TargetShardIntervalsForRouter(Query *query,
-							  RelationRestrictionContext *restrictionContext,
-							  bool *multiShardQuery)
+List *
+TargetShardIntervalsForQuery(Query *query,
+							 RelationRestrictionContext *restrictionContext,
+							 bool *multiShardQuery, Const **partitionValueConst)
 {
 	List *prunedRelationShardList = NIL;
 	ListCell *restrictionCell = NULL;
+	bool multiplePartitionValuesExist = false;
+	Const *queryPartitionValueConst = NULL;
 
 	Assert(restrictionContext != NULL);
 
@@ -1897,18 +2066,45 @@ TargetShardIntervalsForRouter(Query *query,
 		whereFalseQuery = ContainsFalseClause(pseudoRestrictionList);
 		if (!whereFalseQuery && shardCount > 0)
 		{
-			prunedShardList = PruneShards(relationId, tableId, restrictClauseList);
+			Const *restrictionPartitionValueConst = NULL;
+			prunedShardList = PruneShards(relationId, tableId, restrictClauseList,
+										  &restrictionPartitionValueConst);
 
 			if (list_length(prunedShardList) > 1)
 			{
 				(*multiShardQuery) = true;
 			}
+			if (restrictionPartitionValueConst != NULL &&
+				queryPartitionValueConst == NULL)
+			{
+				queryPartitionValueConst = restrictionPartitionValueConst;
+			}
+			else if (restrictionPartitionValueConst != NULL &&
+					 !equal(queryPartitionValueConst, restrictionPartitionValueConst))
+			{
+				multiplePartitionValuesExist = true;
+			}
 		}
 
 		relationRestriction->prunedShardIntervalList = prunedShardList;
 		prunedRelationShardList = lappend(prunedRelationShardList, prunedShardList);
 	}
 
+	/*
+	 * Different resrictions might have different partition columns.
+	 * We report partition column value if there is only one.
+	 */
+	if (multiplePartitionValuesExist)
+	{
+		queryPartitionValueConst = NULL;
+	}
+
+	/* set the outgoing partition column value if requested */
+	if (partitionValueConst != NULL)
+	{
+		*partitionValueConst = queryPartitionValueConst;
+	}
+
 	return prunedRelationShardList;
 }
 
@@ -1950,7 +2146,7 @@ RelationPrunesToMultipleShards(List *relationShardList)
  * exists. The caller should check if there are any shard intervals exist for
  * placement check prior to calling this function.
  */
-static List *
+List *
 WorkersContainingAllShards(List *prunedShardIntervalsList)
 {
 	ListCell *prunedShardIntervalCell = NULL;
@@ -2122,7 +2318,7 @@ BuildRoutesForInsert(Query *query, DeferredErrorMessage **planningError)
 			restrictClauseList = list_make1(equalityExpr);
 
 			prunedShardList = PruneShards(distributedTableId, tableId,
-										  restrictClauseList);
+										  restrictClauseList, NULL);
 		}
 
 		prunedShardCount = list_length(prunedShardList);
@@ -2349,6 +2545,12 @@ IntersectPlacementList(List *lhsPlacementList, List *rhsPlacementList)
 						WORKER_LENGTH) == 0)
 			{
 				placementList = lappend(placementList, rhsPlacement);
+
+				/*
+				 * We don't need to add the same placement over and over again. This
+				 * could happen if both placements of a shard appear on the same node.
+				 */
+				break;
 			}
 		}
 	}
@@ -2483,6 +2685,104 @@ ExtractInsertValuesList(Query *query, Var *partitionColumn)
 }
 
 
+/*
+ * ExtractInsertPartitionKeyValue extracts the partition column value
+ * from an INSERT query. If the expression in the partition column is
+ * non-constant or it is a multi-row INSERT with multiple different partition
+ * column values, the function returns NULL.
+ */
+Const *
+ExtractInsertPartitionKeyValue(Query *query)
+{
+	Oid distributedTableId = ExtractFirstDistributedTableId(query);
+	uint32 rangeTableId = 1;
+	Var *partitionColumn = NULL;
+	TargetEntry *targetEntry = NULL;
+	Const *singlePartitionValueConst = NULL;
+	Node *targetExpression = NULL;
+
+	char partitionMethod = PartitionMethod(distributedTableId);
+	if (partitionMethod == DISTRIBUTE_BY_NONE)
+	{
+		return NULL;
+	}
+
+	partitionColumn = PartitionColumn(distributedTableId, rangeTableId);
+	targetEntry = get_tle_by_resno(query->targetList, partitionColumn->varattno);
+	if (targetEntry == NULL)
+	{
+		/* partition column value not specified */
+		return NULL;
+	}
+
+	targetExpression = strip_implicit_coercions((Node *) targetEntry->expr);
+
+	/*
+	 * Multi-row INSERTs have a Var in the target list that points to
+	 * an RTE_VALUES.
+	 */
+	if (IsA(targetExpression, Var))
+	{
+		Var *partitionVar = (Var *) targetExpression;
+		RangeTblEntry *referencedRTE = NULL;
+		ListCell *valuesListCell = NULL;
+
+		referencedRTE = rt_fetch(partitionVar->varno, query->rtable);
+
+		foreach(valuesListCell, referencedRTE->values_lists)
+		{
+			List *rowValues = (List *) lfirst(valuesListCell);
+			Node *partitionValueNode = list_nth(rowValues, partitionVar->varattno - 1);
+			Expr *partitionValueExpr = (Expr *) strip_implicit_coercions(
+				partitionValueNode);
+			Const *partitionValueConst = NULL;
+
+			if (!IsA(partitionValueExpr, Const))
+			{
+				/* non-constant value in the partition column */
+				singlePartitionValueConst = NULL;
+				break;
+			}
+
+			partitionValueConst = (Const *) partitionValueExpr;
+
+			if (singlePartitionValueConst == NULL)
+			{
+				/* first row has a constant in the partition column, looks promising! */
+				singlePartitionValueConst = partitionValueConst;
+			}
+			else if (!equal(partitionValueConst, singlePartitionValueConst))
+			{
+				/* multiple different values in the partition column, too bad */
+				singlePartitionValueConst = NULL;
+				break;
+			}
+			else
+			{
+				/* another row with the same partition column value! */
+			}
+		}
+	}
+	else if (IsA(targetExpression, Const))
+	{
+		/* single-row INSERT with a constant partition column value */
+		singlePartitionValueConst = (Const *) targetExpression;
+	}
+	else
+	{
+		/* single-row INSERT with a non-constant partition column value */
+		singlePartitionValueConst = NULL;
+	}
+
+	if (singlePartitionValueConst != NULL)
+	{
+		singlePartitionValueConst = copyObject(singlePartitionValueConst);
+	}
+
+	return singlePartitionValueConst;
+}
+
+
 /*
  * MultiRouterPlannableQuery returns true if given query can be router plannable.
  * The query is router plannable if it is a modify query, or if its is a select
@@ -2509,11 +2809,6 @@ MultiRouterPlannableQuery(Query *query, RelationRestrictionContext *restrictionC
 		return false;
 	}
 
-	if (query->hasForUpdate)
-	{
-		return false;
-	}
-
 	foreach(relationRestrictionContextCell, restrictionContext->relationRestrictionList)
 	{
 		RelationRestriction *relationRestriction =
@@ -2537,6 +2832,28 @@ MultiRouterPlannableQuery(Query *query, RelationRestrictionContext *restrictionC
 			{
 				return false;
 			}
+
+			/*
+			 * Currently, we don't support tables with replication factor > 1,
+			 * except reference tables with SELECT ... FOR UDPATE queries. It is
+			 * also not supported from MX nodes.
+			 */
+			if (query->hasForUpdate)
+			{
+				uint32 tableReplicationFactor = TableShardReplicationFactor(
+					distributedTableId);
+
+				if (partitionMethod == DISTRIBUTE_BY_NONE)
+				{
+					EnsureCoordinator();
+				}
+
+
+				if (tableReplicationFactor > 1 && partitionMethod != DISTRIBUTE_BY_NONE)
+				{
+					return false;
+				}
+			}
 		}
 	}
 
