@@ -38,6 +38,9 @@
 #include "distributed/multi_utility.h"
 #include "distributed/pg_dist_partition.h"
 #include "distributed/placement_connection.h"
+#include "distributed/relation_access_tracking.h"
+#include "distributed/query_pushdown_planning.h"
+#include "distributed/query_stats.h"
 #include "distributed/remote_commands.h"
 #include "distributed/shared_library_init.h"
 #include "distributed/statistics_collection.h"
@@ -66,9 +69,15 @@ static void RegisterCitusConfigVariables(void);
 static bool ErrorIfNotASuitableDeadlockFactor(double *newval, void **extra,
 											  GucSource source);
 static void NormalizeWorkerListPath(void);
+static bool NodeConninfoGucCheckHook(char **newval, void **extra, GucSource source);
+static void NodeConninfoGucAssignHook(const char *newval, void *extra);
 static bool StatisticsCollectionGucCheckHook(bool *newval, void **extra, GucSource
 											 source);
 
+/* static variable to hold value of deprecated GUC variable */
+static bool ExpireCachedShards = false;
+static int LargeTableShardCount = 0;
+static int CitusSSLMode = 0;
 
 /* *INDENT-OFF* */
 /* GUC enum definitions */
@@ -110,16 +119,6 @@ static const struct config_enum_entry multi_shard_commit_protocol_options[] = {
 	{ NULL, 0, false }
 };
 
-static const struct config_enum_entry citus_ssl_mode_options[] = {
-	{ "disable", CITUS_SSL_MODE_DISABLE, false },
-	{ "allow", CITUS_SSL_MODE_ALLOW, false },
-	{ "prefer", CITUS_SSL_MODE_PREFER, false },
-	{ "require", CITUS_SSL_MODE_REQUIRE, false },
-	{ "verify-ca", CITUS_SSL_MODE_VERIFY_CA, false },
-	{ "verify-full", CITUS_SSL_MODE_VERIFY_FULL, false },
-	{ NULL, 0, false }
-};
-
 static const struct config_enum_entry multi_task_query_log_level_options[] = {
 	{ "off", MULTI_TASK_QUERY_INFO_OFF, false },
 	{ "debug", DEBUG2, false },
@@ -179,6 +178,8 @@ _PG_init(void)
 		CreateRequiredDirectories();
 	}
 
+	InitConnParams();
+
 	/*
 	 * Register Citus configuration variables. Do so before intercepting
 	 * hooks or calling initialization functions, in case we want to do the
@@ -219,6 +220,7 @@ _PG_init(void)
 	InitializeBackendManagement();
 	InitializeConnectionManagement();
 	InitPlacementConnectionManagement();
+	InitializeCitusQueryStats();
 
 	/* enable modification of pg_catalog tables during pg_upgrade */
 	if (IsBinaryUpgrade)
@@ -322,17 +324,15 @@ RegisterCitusConfigVariables(void)
 		NULL, NULL, NULL);
 	NormalizeWorkerListPath();
 
-	DefineCustomEnumVariable(
+	DefineCustomIntVariable(
 		"citus.sslmode",
-		gettext_noop("SSL mode to use for connections to worker nodes."),
-		gettext_noop("When connecting to a worker node, specify whether the SSL mode"
-					 "mode for the connection is 'disable', 'allow', 'prefer' "
-					 "(the default), 'require', 'verify-ca' or 'verify-full'."),
+		gettext_noop("This variable has been deprecated. Use the citus.node_conninfo "
+					 "GUC instead."),
+		NULL,
 		&CitusSSLMode,
-		CITUS_SSL_MODE_PREFER,
-		citus_ssl_mode_options,
+		0, 0, 32,
 		PGC_POSTMASTER,
-		GUC_SUPERUSER_ONLY,
+		GUC_SUPERUSER_ONLY | GUC_NO_SHOW_ALL,
 		NULL, NULL, NULL);
 
 	DefineCustomBoolVariable(
@@ -360,17 +360,39 @@ RegisterCitusConfigVariables(void)
 
 	DefineCustomBoolVariable(
 		"citus.expire_cached_shards",
-		gettext_noop("Enables shard cache expiration if a shard's size on disk has "
-					 "changed."),
-		gettext_noop("When appending to an existing shard, old data may still be cached "
-					 "on other workers. This configuration entry activates automatic "
-					 "expiration, but should not be used with manual updates to shards."),
+		gettext_noop("This GUC variable has been deprecated."),
+		NULL,
 		&ExpireCachedShards,
 		false,
 		PGC_SIGHUP,
 		0,
 		NULL, NULL, NULL);
 
+	DefineCustomBoolVariable(
+		"citus.enable_single_hash_repartition_joins",
+		gettext_noop("Enables single hash repartitioning between hash "
+					 "distributed tables"),
+		NULL,
+		&EnableSingleHashRepartitioning,
+		false,
+		PGC_USERSET,
+		GUC_NO_SHOW_ALL,
+		NULL, NULL, NULL);
+
+	DefineCustomBoolVariable(
+		"citus.enforce_foreign_key_restrictions",
+		gettext_noop("Enforce restrictions while querying distributed/reference "
+					 "tables with foreign keys"),
+		gettext_noop("When enabled, cascading modifications from reference tables "
+					 "to distributed tables are traced and acted accordingly "
+					 "to avoid creating distributed deadlocks and ensure correctness."),
+		&EnforceForeignKeyRestrictions,
+		true,
+		PGC_USERSET,
+		GUC_NO_SHOW_ALL,
+		NULL, NULL, NULL);
+
+
 	DefineCustomBoolVariable(
 		"citus.subquery_pushdown",
 		gettext_noop("Enables supported subquery pushdown to workers."),
@@ -475,6 +497,23 @@ RegisterCitusConfigVariables(void)
 		GUC_UNIT_MS,
 		NULL, NULL, NULL);
 
+	DefineCustomBoolVariable(
+		"citus.select_opens_transaction_block",
+		gettext_noop("Open transaction blocks for SELECT commands"),
+		gettext_noop("When enabled, Citus will always send a BEGIN to workers when "
+					 "running a distributed SELECT in a transaction block (the "
+					 "default). When disabled, Citus will only send BEGIN before "
+					 "the first write or other operation that requires a distributed "
+					 "transaction, meaning the SELECT on the worker commits "
+					 "immediately, releasing any locks and apply any changes made "
+					 "through function calls even if the distributed transaction "
+					 "aborts."),
+		&SelectOpensTransactionBlock,
+		true,
+		PGC_USERSET,
+		GUC_NO_SHOW_ALL,
+		NULL, NULL, NULL);
+
 	DefineCustomBoolVariable(
 		"citus.enable_deadlock_prevention",
 		gettext_noop("Prevents transactions from expanding to multiple nodes"),
@@ -656,15 +695,12 @@ RegisterCitusConfigVariables(void)
 
 	DefineCustomIntVariable(
 		"citus.large_table_shard_count",
-		gettext_noop("The shard count threshold over which a table is considered large."),
-		gettext_noop("A distributed table is considered to be large if it has "
-					 "more shards than the value specified here. This largeness "
-					 "criteria is then used in picking a table join order during "
-					 "distributed query planning."),
+		gettext_noop("This variable has been deprecated."),
+		gettext_noop("Consider reference tables instead"),
 		&LargeTableShardCount,
 		4, 1, 10000,
 		PGC_USERSET,
-		0,
+		GUC_NO_SHOW_ALL,
 		NULL, NULL, NULL);
 
 	DefineCustomIntVariable(
@@ -916,6 +952,20 @@ RegisterCitusConfigVariables(void)
 		&StatisticsCollectionGucCheckHook,
 		NULL, NULL);
 
+	DefineCustomStringVariable(
+		"citus.node_conninfo",
+		gettext_noop("Sets parameters used for outbound connections."),
+		NULL,
+		&NodeConninfo,
+		"sslmode=prefer",
+		PGC_POSTMASTER,
+		GUC_SUPERUSER_ONLY,
+		NodeConninfoGucCheckHook,
+		NodeConninfoGucAssignHook,
+		NULL);
+	NormalizeWorkerListPath();
+
+
 	/* warn about config items in the citus namespace that are not registered above */
 	EmitWarningsOnPlaceholders("citus");
 }
@@ -987,6 +1037,87 @@ NormalizeWorkerListPath(void)
 }
 
 
+/*
+ * NodeConninfoGucCheckHook ensures conninfo settings are in the expected form
+ * and that the keywords of all non-null settings are on a whitelist devised to
+ * keep users from setting options that may result in confusion.
+ */
+static bool
+NodeConninfoGucCheckHook(char **newval, void **extra, GucSource source)
+{
+	/* this array _must_ be kept in an order usable by bsearch */
+	const char *whitelist[] = {
+		"application_name",
+		"connect_timeout",
+			#if defined(ENABLE_GSS) && defined(ENABLE_SSPI)
+		"gsslib",
+			#endif
+		"keepalives",
+		"keepalives_count",
+		"keepalives_idle",
+		"keepalives_interval",
+			#if defined(ENABLE_GSS) || defined(ENABLE_SSPI)
+		"krbsrvname",
+			#endif
+		"sslcompression",
+		"sslcrl",
+		"sslmode",
+		"sslrootcert"
+	};
+	char *errorMsg = NULL;
+	bool conninfoValid = CheckConninfo(*newval, whitelist, lengthof(whitelist),
+									   &errorMsg);
+
+	if (!conninfoValid)
+	{
+		GUC_check_errdetail("%s", errorMsg);
+	}
+
+	return conninfoValid;
+}
+
+
+/*
+ * NodeConninfoGucAssignHook is the assignment hook for the node_conninfo GUC
+ * variable. Though this GUC is a "string", we actually parse it as a non-URI
+ * PQconninfo key/value setting, storing the resultant PQconninfoOption values
+ * using the public functions in connection_configuration.c.
+ */
+static void
+NodeConninfoGucAssignHook(const char *newval, void *extra)
+{
+	PQconninfoOption *optionArray = NULL;
+	PQconninfoOption *option = NULL;
+
+	if (newval == NULL)
+	{
+		newval = "";
+	}
+
+	optionArray = PQconninfoParse(newval, NULL);
+	if (optionArray == NULL)
+	{
+		ereport(FATAL, (errmsg("cannot parse node_conninfo value"),
+						errdetail("The GUC check hook should prevent "
+								  "all malformed values.")));
+	}
+
+	ResetConnParams();
+
+	for (option = optionArray; option->keyword != NULL; option++)
+	{
+		if (option->val == NULL || option->val[0] == '\0')
+		{
+			continue;
+		}
+
+		AddConnParam(option->keyword, option->val);
+	}
+
+	PQconninfoFree(optionArray);
+}
+
+
 static bool
 StatisticsCollectionGucCheckHook(bool *newval, void **extra, GucSource source)
 {
