@@ -17,6 +17,7 @@
 #include "access/hash.h"
 #include "distributed/backend_data.h"
 #include "distributed/distributed_deadlock_detection.h"
+#include "distributed/errormessage.h"
 #include "distributed/hash_helpers.h"
 #include "distributed/listutils.h"
 #include "distributed/lock_graph.h"
@@ -79,7 +80,7 @@ check_distributed_deadlocks(PG_FUNCTION_ARGS)
 
 
 /*
- * CheckForDistributedDeadlocks is the entry point for detecing
+ * CheckForDistributedDeadlocks is the entry point for detecting
  * distributed deadlocks.
  *
  * In plain words, the function first builds a wait graph by
@@ -673,7 +674,7 @@ LogDistributedDeadlockDebugMessage(const char *errorMessage)
 	}
 
 	ereport(LOG, (errmsg("[%s] %s", timestamptz_to_str(GetCurrentTimestamp()),
-						 errorMessage)));
+						 ApplyLogRedaction(errorMessage))));
 }
 
 
