@@ -81,7 +81,7 @@ OpenTransactionsForAllTasks(List *taskList, int connectionFlags)
 		else
 		{
 			/* can only open connections for DDL and DML commands */
-			Assert(task->taskType == DDL_TASK);
+			Assert(task->taskType == DDL_TASK || VACUUM_ANALYZE_TASK);
 
 			accessType = PLACEMENT_ACCESS_DDL;
 		}
@@ -91,7 +91,6 @@ OpenTransactionsForAllTasks(List *taskList, int connectionFlags)
 			ShardPlacement *shardPlacement = (ShardPlacement *) lfirst(placementCell);
 			ShardPlacementAccess placementModification;
 			List *placementAccessList = NIL;
-			List *placementSelectList = NIL;
 			MultiConnection *connection = NULL;
 
 			WorkerNode *workerNode = FindWorkerNode(shardPlacement->nodeName,
@@ -109,10 +108,27 @@ OpenTransactionsForAllTasks(List *taskList, int connectionFlags)
 
 			placementAccessList = lappend(placementAccessList, &placementModification);
 
-			/* add additional placement accesses for subselects (e.g. INSERT .. SELECT) */
-			placementSelectList = BuildPlacementSelectList(shardPlacement->groupId,
-														   task->relationShardList);
-			placementAccessList = list_concat(placementAccessList, placementSelectList);
+			if (accessType == PLACEMENT_ACCESS_DDL)
+			{
+				List *placementDDLList = BuildPlacementDDLList(shardPlacement->groupId,
+															   task->relationShardList);
+
+				/*
+				 * All relations appearing inter-shard DDL commands should be marked
+				 * with DDL access.
+				 */
+				placementAccessList = list_concat(placementAccessList, placementDDLList);
+			}
+			else
+			{
+				List *placementSelectList =
+					BuildPlacementSelectList(shardPlacement->groupId,
+											 task->relationShardList);
+
+				/* add additional placement accesses for subselects (e.g. INSERT .. SELECT) */
+				placementAccessList =
+					list_concat(placementAccessList, placementSelectList);
+			}
 
 			/*
 			 * Find a connection that sees preceding writes and cannot self-deadlock,
