@@ -43,7 +43,6 @@ static void StartRemoteTransactionSavepointRollback(MultiConnection *connection,
 static void FinishRemoteTransactionSavepointRollback(MultiConnection *connection,
 													 SubTransactionId subId);
 
-static void CheckTransactionHealth(void);
 static void Assign2PCIdentifier(MultiConnection *connection);
 static void WarnAboutLeakedPreparedTransaction(MultiConnection *connection, bool commit);
 
@@ -107,8 +106,9 @@ StartRemoteTransactionBegin(struct MultiConnection *connection)
 
 	if (!SendRemoteCommand(connection, beginAndSetDistributedTransactionId->data))
 	{
-		ReportConnectionError(connection, WARNING);
-		MarkRemoteTransactionFailed(connection, true);
+		const bool raiseErrors = true;
+
+		HandleRemoteTransactionConnectionError(connection, raiseErrors);
 	}
 }
 
@@ -187,7 +187,7 @@ void
 StartRemoteTransactionCommit(MultiConnection *connection)
 {
 	RemoteTransaction *transaction = &connection->remoteTransaction;
-	const bool dontRaiseError = false;
+	const bool raiseErrors = false;
 	const bool isCommit = true;
 
 	/* can only commit if transaction is in progress */
@@ -225,8 +225,7 @@ StartRemoteTransactionCommit(MultiConnection *connection)
 
 		if (!SendRemoteCommand(connection, command.data))
 		{
-			ReportConnectionError(connection, WARNING);
-			MarkRemoteTransactionFailed(connection, dontRaiseError);
+			HandleRemoteTransactionConnectionError(connection, raiseErrors);
 
 			WarnAboutLeakedPreparedTransaction(connection, isCommit);
 		}
@@ -244,8 +243,7 @@ StartRemoteTransactionCommit(MultiConnection *connection)
 			 * Failing in this state means that we don't know whether the the
 			 * commit has succeeded.
 			 */
-			ReportConnectionError(connection, WARNING);
-			MarkRemoteTransactionFailed(connection, dontRaiseError);
+			HandleRemoteTransactionConnectionError(connection, raiseErrors);
 		}
 	}
 }
@@ -261,19 +259,18 @@ FinishRemoteTransactionCommit(MultiConnection *connection)
 {
 	RemoteTransaction *transaction = &connection->remoteTransaction;
 	PGresult *result = NULL;
-	const bool dontRaiseErrors = false;
+	const bool raiseErrors = false;
 	const bool isCommit = true;
 
 	Assert(transaction->transactionState == REMOTE_TRANS_1PC_ABORTING ||
 		   transaction->transactionState == REMOTE_TRANS_1PC_COMMITTING ||
 		   transaction->transactionState == REMOTE_TRANS_2PC_COMMITTING);
 
-	result = GetRemoteCommandResult(connection, dontRaiseErrors);
+	result = GetRemoteCommandResult(connection, raiseErrors);
 
 	if (!IsResponseOK(result))
 	{
-		ReportResultError(connection, result, WARNING);
-		MarkRemoteTransactionFailed(connection, dontRaiseErrors);
+		HandleRemoteTransactionResultError(connection, result, raiseErrors);
 
 		/*
 		 * Failing in this state means that we will often not know whether
@@ -343,7 +340,7 @@ void
 StartRemoteTransactionAbort(MultiConnection *connection)
 {
 	RemoteTransaction *transaction = &connection->remoteTransaction;
-	const bool dontRaiseErrors = false;
+	const bool raiseErrors = false;
 	const bool isNotCommit = false;
 
 	Assert(transaction->transactionState != REMOTE_TRANS_INVALID);
@@ -370,8 +367,7 @@ StartRemoteTransactionAbort(MultiConnection *connection)
 
 		if (!SendRemoteCommand(connection, command.data))
 		{
-			ReportConnectionError(connection, WARNING);
-			MarkRemoteTransactionFailed(connection, dontRaiseErrors);
+			HandleRemoteTransactionConnectionError(connection, raiseErrors);
 
 			WarnAboutLeakedPreparedTransaction(connection, isNotCommit);
 		}
@@ -400,7 +396,7 @@ StartRemoteTransactionAbort(MultiConnection *connection)
 		if (!SendRemoteCommand(connection, "ROLLBACK"))
 		{
 			/* no point in reporting a likely redundant message */
-			MarkRemoteTransactionFailed(connection, dontRaiseErrors);
+			MarkRemoteTransactionFailed(connection, raiseErrors);
 		}
 		else
 		{
@@ -418,34 +414,33 @@ void
 FinishRemoteTransactionAbort(MultiConnection *connection)
 {
 	RemoteTransaction *transaction = &connection->remoteTransaction;
-	PGresult *result = NULL;
-	const bool dontRaiseErrors = false;
-	const bool isNotCommit = false;
+	const bool raiseErrors = false;
 
-	result = GetRemoteCommandResult(connection, dontRaiseErrors);
-
-	if (!IsResponseOK(result))
+	if (transaction->transactionState == REMOTE_TRANS_2PC_ABORTING)
 	{
-		ReportResultError(connection, result, WARNING);
-		MarkRemoteTransactionFailed(connection, dontRaiseErrors);
+		PGresult *result = GetRemoteCommandResult(connection, raiseErrors);
+		if (!IsResponseOK(result))
+		{
+			const bool isCommit = false;
 
-		if (transaction->transactionState == REMOTE_TRANS_2PC_ABORTING)
-		{
-			WarnAboutLeakedPreparedTransaction(connection, isNotCommit);
-		}
-		else
-		{
-			ereport(WARNING,
-					(errmsg("failed to abort 1PC transaction \"%s\" on %s:%d",
-							transaction->preparedName, connection->hostname,
-							connection->port)));
+			HandleRemoteTransactionResultError(connection, result, raiseErrors);
+
+			WarnAboutLeakedPreparedTransaction(connection, isCommit);
 		}
+
+		PQclear(result);
 	}
 
-	PQclear(result);
-
-	result = GetRemoteCommandResult(connection, dontRaiseErrors);
-	Assert(!result);
+	/*
+	 * Try to consume results of any in-progress commands. In the 1PC case
+	 * this is also where we consume the result of the ROLLBACK.
+	 *
+	 * If we don't succeed the connection will be in a bad state, so we close it.
+	 */
+	if (!ClearResults(connection, raiseErrors))
+	{
+		ShutdownConnection(connection);
+	}
 
 	transaction->transactionState = REMOTE_TRANS_ABORTED;
 }
@@ -498,8 +493,7 @@ StartRemoteTransactionPrepare(struct MultiConnection *connection)
 
 	if (!SendRemoteCommand(connection, command.data))
 	{
-		ReportConnectionError(connection, WARNING);
-		MarkRemoteTransactionFailed(connection, raiseErrors);
+		HandleRemoteTransactionConnectionError(connection, raiseErrors);
 	}
 	else
 	{
@@ -526,9 +520,8 @@ FinishRemoteTransactionPrepare(struct MultiConnection *connection)
 
 	if (!IsResponseOK(result))
 	{
-		ReportResultError(connection, result, WARNING);
 		transaction->transactionState = REMOTE_TRANS_ABORTED;
-		MarkRemoteTransactionFailed(connection, raiseErrors);
+		HandleRemoteTransactionResultError(connection, result, raiseErrors);
 	}
 	else
 	{
@@ -639,6 +632,53 @@ RemoteTransactionsBeginIfNecessary(List *connectionList)
 }
 
 
+/*
+ * HandleRemoteTransactionConnectionError records a transaction as having failed
+ * and throws a connection error if the transaction was critical and raiseErrors
+ * is true, or a warning otherwise.
+ */
+void
+HandleRemoteTransactionConnectionError(MultiConnection *connection, bool raiseErrors)
+{
+	RemoteTransaction *transaction = &connection->remoteTransaction;
+
+	transaction->transactionFailed = true;
+
+	if (transaction->transactionCritical && raiseErrors)
+	{
+		ReportConnectionError(connection, ERROR);
+	}
+	else
+	{
+		ReportConnectionError(connection, WARNING);
+	}
+}
+
+
+/*
+ * HandleRemoteTransactionResultError records a transaction as having failed
+ * and throws a result error if the transaction was critical and raiseErrors
+ * is true, or a warning otherwise.
+ */
+void
+HandleRemoteTransactionResultError(MultiConnection *connection, PGresult *result, bool
+								   raiseErrors)
+{
+	RemoteTransaction *transaction = &connection->remoteTransaction;
+
+	transaction->transactionFailed = true;
+
+	if (transaction->transactionCritical && raiseErrors)
+	{
+		ReportResultError(connection, result, ERROR);
+	}
+	else
+	{
+		ReportResultError(connection, result, WARNING);
+	}
+}
+
+
 /*
  * MarkRemoteTransactionFailed records a transaction as having failed.
  *
@@ -680,6 +720,19 @@ MarkRemoteTransactionCritical(struct MultiConnection *connection)
 }
 
 
+/*
+ * IsRemoteTransactionCritical returns whether the remote transaction on
+ * the given connection has been marked as critical.
+ */
+bool
+IsRemoteTransactionCritical(struct MultiConnection *connection)
+{
+	RemoteTransaction *transaction = &connection->remoteTransaction;
+
+	return transaction->transactionCritical;
+}
+
+
 /*
  * CloseRemoteTransaction handles closing a connection that, potentially, is
  * part of a coordinated transaction.  This should only ever be called from
@@ -783,12 +836,6 @@ CoordinatedRemoteTransactionsCommit(void)
 	List *connectionList = NIL;
 	bool raiseInterrupts = false;
 
-	/*
-	 * Before starting to commit on any of the nodes - after which we can't
-	 * completely roll-back anymore - check that things are in a good state.
-	 */
-	CheckTransactionHealth();
-
 	/*
 	 * Issue appropriate transaction commands to remote nodes. If everything
 	 * went well that's going to be COMMIT or COMMIT PREPARED, if individual
@@ -997,7 +1044,7 @@ void
 CoordinatedRemoteTransactionsSavepointRollback(SubTransactionId subId)
 {
 	dlist_iter iter;
-	const bool dontRaiseInterrupts = false;
+	const bool raiseInterrupts = false;
 	List *connectionList = NIL;
 
 	/* asynchronously send ROLLBACK TO SAVEPOINT */
@@ -1028,7 +1075,7 @@ CoordinatedRemoteTransactionsSavepointRollback(SubTransactionId subId)
 		connectionList = lappend(connectionList, connection);
 	}
 
-	WaitForAllConnections(connectionList, dontRaiseInterrupts);
+	WaitForAllConnections(connectionList, raiseInterrupts);
 
 	/* and wait for the results */
 	dlist_foreach(iter, &InProgressTransactions)
@@ -1059,8 +1106,7 @@ StartRemoteTransactionSavepointBegin(MultiConnection *connection, SubTransaction
 
 	if (!SendRemoteCommand(connection, savepointCommand->data))
 	{
-		ReportConnectionError(connection, WARNING);
-		MarkRemoteTransactionFailed(connection, raiseErrors);
+		HandleRemoteTransactionConnectionError(connection, raiseErrors);
 	}
 }
 
@@ -1077,8 +1123,7 @@ FinishRemoteTransactionSavepointBegin(MultiConnection *connection, SubTransactio
 	PGresult *result = GetRemoteCommandResult(connection, raiseErrors);
 	if (!IsResponseOK(result))
 	{
-		ReportResultError(connection, result, WARNING);
-		MarkRemoteTransactionFailed(connection, raiseErrors);
+		HandleRemoteTransactionResultError(connection, result, raiseErrors);
 	}
 
 	PQclear(result);
@@ -1100,8 +1145,7 @@ StartRemoteTransactionSavepointRelease(MultiConnection *connection,
 
 	if (!SendRemoteCommand(connection, savepointCommand->data))
 	{
-		ReportConnectionError(connection, WARNING);
-		MarkRemoteTransactionFailed(connection, raiseErrors);
+		HandleRemoteTransactionConnectionError(connection, raiseErrors);
 	}
 }
 
@@ -1119,8 +1163,7 @@ FinishRemoteTransactionSavepointRelease(MultiConnection *connection,
 	PGresult *result = GetRemoteCommandResult(connection, raiseErrors);
 	if (!IsResponseOK(result))
 	{
-		ReportResultError(connection, result, WARNING);
-		MarkRemoteTransactionFailed(connection, raiseErrors);
+		HandleRemoteTransactionResultError(connection, result, raiseErrors);
 	}
 
 	PQclear(result);
@@ -1136,14 +1179,13 @@ static void
 StartRemoteTransactionSavepointRollback(MultiConnection *connection,
 										SubTransactionId subId)
 {
-	const bool dontRaiseErrors = false;
+	const bool raiseErrors = false;
 	StringInfo savepointCommand = makeStringInfo();
 	appendStringInfo(savepointCommand, "ROLLBACK TO SAVEPOINT savepoint_%u", subId);
 
 	if (!SendRemoteCommand(connection, savepointCommand->data))
 	{
-		ReportConnectionError(connection, WARNING);
-		MarkRemoteTransactionFailed(connection, dontRaiseErrors);
+		HandleRemoteTransactionConnectionError(connection, raiseErrors);
 	}
 }
 
@@ -1158,14 +1200,13 @@ static void
 FinishRemoteTransactionSavepointRollback(MultiConnection *connection, SubTransactionId
 										 subId)
 {
-	const bool dontRaiseErrors = false;
+	const bool raiseErrors = false;
 	RemoteTransaction *transaction = &connection->remoteTransaction;
 
-	PGresult *result = GetRemoteCommandResult(connection, dontRaiseErrors);
+	PGresult *result = GetRemoteCommandResult(connection, raiseErrors);
 	if (!IsResponseOK(result))
 	{
-		ReportResultError(connection, result, WARNING);
-		MarkRemoteTransactionFailed(connection, dontRaiseErrors);
+		HandleRemoteTransactionResultError(connection, result, raiseErrors);
 	}
 
 	/* ROLLBACK TO SAVEPOINT succeeded, check if it recovers the transaction */
@@ -1181,13 +1222,13 @@ FinishRemoteTransactionSavepointRollback(MultiConnection *connection, SubTransac
 
 
 /*
- * CheckTransactionHealth checks if any of the participating transactions in a
+ * CheckRemoteTransactionsHealth checks if any of the participating transactions in a
  * coordinated transaction failed, and what consequence that should have.
  * This needs to be called before the coordinated transaction commits (but
  * after they've been PREPAREd if 2PC is in use).
  */
-static void
-CheckTransactionHealth(void)
+void
+CheckRemoteTransactionsHealth(void)
 {
 	dlist_iter iter;
 
