@@ -16,7 +16,11 @@ use warnings;
 
 use Fcntl;
 use Getopt::Long;
-
+use File::Spec::Functions;
+use File::Path qw(make_path remove_tree);
+use Config;
+use POSIX qw( WNOHANG mkfifo );
+use Cwd 'abs_path';
 
 sub Usage()
 {
@@ -40,6 +44,7 @@ sub Usage()
     print "  --valgrind-log-file	Path to the write valgrind logs\n";
     print "  --pg_ctl-timeout    	Timeout for pg_ctl\n";
     print "  --connection-timeout	Timeout for connecting to worker nodes\n";
+    print "  --mitmproxy        	Start a mitmproxy for one of the workers\n";
     exit 1;
 }
 
@@ -65,8 +70,17 @@ my $valgrindPath = "valgrind";
 my $valgrindLogFile = "valgrind_test_log.txt";
 my $pgCtlTimeout = undef;
 my $connectionTimeout = 5000;
+my $useMitmproxy = 0;
+my $mitmFifoPath = catfile("tmp_check", "mitmproxy.fifo");
 
 my $serversAreShutdown = "TRUE";
+my $usingWindows = 0;
+my $mitmPid = 0;
+
+if ($Config{osname} eq "MSWin32")
+{
+	$usingWindows = 1;
+};
 
 GetOptions(
     'isolationtester' => \$isolationtester,
@@ -85,6 +99,7 @@ GetOptions(
     'valgrind-log-file=s' => \$valgrindLogFile,
     'pg_ctl-timeout=s' => \$pgCtlTimeout,
     'connection-timeout=s' => \$connectionTimeout,
+    'mitmproxy' => \$useMitmproxy,
     'help' => sub { Usage() });
 
 # Update environment to include [DY]LD_LIBRARY_PATH/LIBDIR/etc -
@@ -99,7 +114,11 @@ if (defined $libdir)
     $ENV{LD_LIBRARY_PATH} = "$libdir:".($ENV{LD_LIBRARY_PATH} || '');
     $ENV{DYLD_LIBRARY_PATH} = "$libdir:".($ENV{DYLD_LIBRARY_PATH} || '');
     $ENV{LIBPATH} = "$libdir:".($ENV{LIBPATH} || '');
-    $ENV{PATH} = "$libdir:".($ENV{PATH} || '');
+    if ($usingWindows) {
+        $ENV{PATH} = "$libdir;".($ENV{PATH} || '');
+    } else {
+        $ENV{PATH} = "$libdir:".($ENV{PATH} || '');
+    }
 }
 
 # Put $bindir to the end of PATH. We want to prefer system binaries by
@@ -107,18 +126,38 @@ if (defined $libdir)
 # want to find binaries if they're not in PATH.
 if (defined $bindir)
 {
-    $ENV{PATH} = ($ENV{PATH} || '').":$bindir";
+    if ($usingWindows) {
+        $ENV{PATH} = ($ENV{PATH} || '').";$bindir";
+    } else {
+        $ENV{PATH} = ($ENV{PATH} || '').":$bindir";
+    }
 }
 
-
 # Most people are used to unified diffs these days, rather than the
 # context diffs pg_regress defaults to.  Change default to avoid
 # everyone having to (re-)learn how to change that setting.  Also add
 # a bit more context to make it easier to locate failed test sections.
-$ENV{PG_REGRESS_DIFF_OPTS} = '-dU10';
+#
+# Also, ignore whitespace, without this the diffs on windows are unreadable
+$ENV{PG_REGRESS_DIFF_OPTS} = '-dU10 -w';
+
+my $plainRegress = "";
+my $isolationRegress = "";
+my $pgConfig = "";
+
+if ($usingWindows)
+{
+	$plainRegress = "$bindir\\pg_regress.exe";
+	$isolationRegress = "$bindir\\pg_isolation_regress.exe";
+	$pgConfig = "$bindir\\pg_config.exe";
+}
+else
+{
+	$plainRegress = "$pgxsdir/src/test/regress/pg_regress";
+	$isolationRegress = "${postgresBuilddir}/src/test/isolation/pg_isolation_regress";
+	$pgConfig = "$bindir/pg_config";
+}
 
-my $plainRegress = "$pgxsdir/src/test/regress/pg_regress";
-my $isolationRegress = "${postgresBuilddir}/src/test/isolation/pg_isolation_regress";
 if ($isolationtester && ! -f "$isolationRegress")
 {
     die <<"MESSAGE";
@@ -136,7 +175,7 @@ directory is present locally
 MESSAGE
 }
 
-my $vanillaRegress = "${postgresBuilddir}/src/test/regress/pg_regress";
+my $vanillaRegress = catfile("${postgresBuilddir}", "src", "test", "regress", "pg_regress");
 if ($vanillatest && ! -f "$vanillaRegress")
 {
     die <<"MESSAGE";
@@ -149,6 +188,11 @@ are present.
 MESSAGE
 }
 
+if ($useMitmproxy)
+{
+  system("mitmdump --version") == 0 or die "make sure mitmdump is on PATH";
+}
+
 # If pgCtlTimeout is defined, we will set related environment variable.
 # This is generally used with valgrind because valgrind starts slow and we
 # need to increase timeout.
@@ -164,18 +208,18 @@ if (defined $pgCtlTimeout)
 # only reliable way to do this.
 sub replace_postgres
 {
-    if (-e "$bindir/postgres.orig")
+    if (-e catfile("$bindir", "postgres.orig"))
     {
 	print "wrapper exists\n";
     }
     else
     {
 	print "moving $bindir/postgres to $bindir/postgres.orig\n";
-	rename "$bindir/postgres", "$bindir/postgres.orig"
+	rename catfile("$bindir", "postgres"), catfile("$bindir", "postgres.orig")
 	    or die "Could not move postgres out of the way";
     }
 
-    sysopen my $fh, "$bindir/postgres", O_CREAT|O_TRUNC|O_RDWR, 0700
+    sysopen my $fh, catfile("$bindir", "postgres"), O_CREAT|O_TRUNC|O_RDWR, 0700
 	or die "Could not create postgres wrapper at $bindir/postgres";
     print $fh <<"END";
 #!/bin/bash
@@ -195,11 +239,11 @@ END
 # revert changes replace_postgres() performed
 sub revert_replace_postgres
 {
-    if (-e "$bindir/postgres.orig")
+    if (-e catfile("$bindir", "postgres.orig"))
     {
 	print "wrapper exists, removing\n";
 	print "moving $bindir/postgres.orig to $bindir/postgres\n";
-	rename "$bindir/postgres.orig", "$bindir/postgres"
+	rename catfile("$bindir", "postgres.orig"), catfile("$bindir", "postgres")
 	    or die "Could not move postgres back";
     }
 }
@@ -230,16 +274,29 @@ my $user = "postgres";
 my @pgOptions = ();
 
 # Postgres options set for the tests
-push(@pgOptions, '-c', "listen_addresses='${host}'");
+push(@pgOptions, '-c', "listen_addresses=${host}");
 # not required, and we don't necessarily have access to the default directory
 push(@pgOptions, '-c', "unix_socket_directories=");
 push(@pgOptions, '-c', "fsync=off");
-push(@pgOptions, '-c', "shared_preload_libraries=citus");
+
+my $sharedPreloadLibraries = "citus";
+
+# check if pg_stat_statements extension is installed
+# if it is add it to shared preload libraries
+my $sharedir = `$pgConfig --sharedir`;
+chomp $sharedir;
+my $pg_stat_statements_control = catfile($sharedir, "extension", "pg_stat_statements.control");
+if (-e $pg_stat_statements_control)
+{
+	$sharedPreloadLibraries .= ',pg_stat_statements'; 
+}
+push(@pgOptions, '-c', "shared_preload_libraries=${sharedPreloadLibraries}");
+
 push(@pgOptions, '-c', "wal_level=logical");
 
 # Citus options set for the tests
 push(@pgOptions, '-c', "citus.shard_count=4");
-push(@pgOptions, '-c', "citus.shard_max_size=300kB");
+push(@pgOptions, '-c', "citus.shard_max_size=1500kB");
 push(@pgOptions, '-c', "citus.max_running_tasks_per_node=4");
 push(@pgOptions, '-c', "citus.expire_cached_shards=on");
 push(@pgOptions, '-c', "citus.task_tracker_delay=10ms");
@@ -247,6 +304,23 @@ push(@pgOptions, '-c', "citus.remote_task_check_interval=1ms");
 push(@pgOptions, '-c', "citus.shard_replication_factor=2");
 push(@pgOptions, '-c', "citus.node_connection_timeout=${connectionTimeout}");
 
+if ($useMitmproxy)
+{
+  # make tests reproducible by never trying to negotiate ssl
+  push(@pgOptions, '-c', "citus.node_conninfo=sslmode=disable");
+}
+
+if ($useMitmproxy)
+{
+  if (! -e "tmp_check")
+  {
+    make_path("tmp_check") or die 'could not create tmp_check directory';
+  }
+  my $absoluteFifoPath = abs_path($mitmFifoPath);
+  die 'abs_path returned empty string' unless ($absoluteFifoPath ne "");
+  push(@pgOptions, '-c', "citus.mitmfifo=$absoluteFifoPath");
+}
+
 if ($followercluster)
 {
   push(@pgOptions, '-c', "max_wal_senders=10");
@@ -295,27 +369,53 @@ for my $option (@userPgOptions)
 %fdwServers = ('fake_fdw_server', 'fake_fdw');
 
 # Cleanup leftovers and prepare directories for the run
-system("rm", ('-rf', 'tmp_check/tmp-bin')) == 0 or die "Could not remove tmp-bin directory";
-
-system("rm", ('-rf', 'tmp_check/master')) == 0 or die "Could not remove master directory";
-for my $port (@workerPorts)
+if (-e catfile('tmp_check', 'tmp-bin'))
 {
-    system("rm", ('-rf', "tmp_check/worker.$port")) == 0 or die "Could not remove worker directory";
+	remove_tree(catfile('tmp_check', 'tmp-bin')) or die "Could not remove tmp-bin directory";
+}
+
+if (-e catfile('tmp_check', 'master'))
+{
+	remove_tree(catfile('tmp_check', 'master')) or die "Could not remove master directory";
+}
+
+for my $port (@workerPorts)
+{
+	if (-e catfile("tmp_check", "worker.$port"))
+	{
+    		remove_tree(catfile("tmp_check", "worker.$port")) or die "Could not remove worker directory";
+	}
+}
+
+if (-e catfile("tmp_check", "master-follower"))
+{
+	remove_tree(catfile("tmp_check", "master-follower")) or die "Could not remove master directory";
 }
 
-system("rm", ('-rf', 'tmp_check/master-follower')) == 0 or die "Could not remove master directory";
 for my $port (@followerWorkerPorts)
 {
-    system("rm", ('-rf', "tmp_check/follower.$port")) == 0 or die "Could not remove worker directory";
+	if (-e catfile("tmp_check", "follower.$port"))
+	{
+	    remove_tree(catfile("tmp_check", "follower.$port")) or die "Could not remove worker directory";
+	}
 }
 
 # Prepare directory in which 'psql' has some helpful variables for locating the workers
-system("mkdir", ('-p', "tmp_check/tmp-bin")) == 0
-	or die "Could not create tmp-bin directory";
-sysopen my $fh, "tmp_check/tmp-bin/psql", O_CREAT|O_TRUNC|O_RDWR, 0700
+make_path(catfile("tmp_check", "tmp-bin")) or die "Could not create tmp_bin directory $!\n";
+
+my $psql_name = "psql";
+if ($usingWindows)
+{
+	$psql_name = "psql.cmd";
+}
+
+sysopen my $fh, catfile("tmp_check", "tmp-bin", $psql_name), O_CREAT|O_TRUNC|O_RDWR, 0700
 	or die "Could not create psql wrapper";
-print $fh "#!/bin/bash\n";
-print $fh "exec psql ";
+if ($usingWindows)
+{ 
+    print $fh "\@echo off\n";
+}
+print $fh catfile($bindir, "psql")." ";
 print $fh "--variable=master_port=$masterPort ";
 print $fh "--variable=follower_master_port=$followerCoordPort ";
 print $fh "--variable=default_user=$user ";
@@ -330,43 +430,81 @@ for my $workeroff (0 .. $#followerWorkerPorts)
 	my $port = $followerWorkerPorts[$workeroff];
 	print $fh "--variable=follower_worker_".($workeroff+1)."_port=$port ";
 }
-print $fh "\"\$@\"\n"; # pass on the commandline arguments
+
+if ($usingWindows)
+{
+	print $fh "--variable=dev_null=\"/nul\" ";
+	print $fh "--variable=temp_dir=\"%TEMP%\" ";
+}
+else
+{
+	print $fh "--variable=dev_null=\"/dev/null\" ";	
+	print $fh "--variable=temp_dir=\"/tmp/\" ";
+}
+
+
+if ($usingWindows)
+{
+	print $fh " %*\n"; # pass on the commandline arguments
+}
+else
+{
+	print $fh " \"\$@\"\n"; # pass on the commandline arguments
+}
 close $fh;
 
-system("mkdir", ('-p', 'tmp_check/master/log')) == 0 or die "Could not create master directory";
+make_path(catfile('tmp_check', 'master', 'log')) or die 'Could not create master directory';
 for my $port (@workerPorts)
 {
-    system("mkdir", ('-p', "tmp_check/worker.$port/log")) == 0
+    make_path(catfile("tmp_check", "worker.$port", "log"))
         or die "Could not create worker directory";
 }
 
 if ($followercluster)
 {
-    system("mkdir", ('-p', 'tmp_check/master-follower/log')) == 0 or die "Could not create follower directory";
+    make_path(catfile('tmp_check', 'master-follower', 'log')) or die "Could not create follower directory";
     for my $port (@followerWorkerPorts)
     {
-        system("mkdir", ('-p', "tmp_check/follower.$port/log")) == 0
+        make_path(catfile("tmp_check", "follower.$port", "log"))
             or die "Could not create worker directory";
     }
 }
 
 # Create new data directories, copy workers for speed
-system("$bindir/initdb", ("--nosync", "-U", $user, "tmp_check/master/data")) == 0
+system(catfile("$bindir", "initdb"), ("--nosync", "-U", $user, "--encoding", "UTF8", catfile("tmp_check", "master", "data"))) == 0
     or die "Could not create master data directory";
 
 if ($followercluster)
 {
   # This is only necessary on PG 9.6 but it doesn't hurt PG 10
-  open(my $fd, ">>", "tmp_check/master/data/pg_hba.conf")
+  open(my $fd, ">>", catfile("tmp_check", "master", "data", "pg_hba.conf"))
     or die "could not open pg_hba.conf";
   print $fd "\nhost replication postgres 127.0.0.1/32 trust";
   close $fd;
 }
 
-for my $port (@workerPorts)
+if ($usingWindows)
 {
-    system("cp -a tmp_check/master/data tmp_check/worker.$port/data") == 0
-        or die "Could not create worker data directory";
+	for my $port (@workerPorts)
+	{
+		system(catfile("$bindir", "initdb"), ("--nosync", "-U", $user, "--encoding", "UTF8", catfile("tmp_check", "worker.$port", "data"))) == 0
+		    or die "Could not create worker data directory";
+	}	
+}
+else
+{
+	for my $port (@workerPorts)
+	{
+	    system("cp", ("-a", catfile("tmp_check", "master", "data"), catfile("tmp_check", "worker.$port", "data"))) == 0
+	        or die "Could not create worker data directory";
+	}
+}
+
+if ($usingWindows)
+{
+	# takeown returns a failing result code no matter what happens, so skip the check
+	system("takeown /f data");
+	system("icacls data /grant \%userdomain\%\\\%username\%:F");
 }
 
 # Routine to shutdown servers at failure/exit
@@ -374,34 +512,72 @@ sub ShutdownServers()
 {
     if ($serversAreShutdown eq "FALSE")
     {
-        system("$bindir/pg_ctl",
-               ('stop', '-w', '-D', 'tmp_check/master/data')) == 0
+        system(catfile("$bindir", "pg_ctl"),
+               ('stop', '-w', '-D', catfile('tmp_check', 'master', 'data'))) == 0
             or warn "Could not shutdown worker server";
 
         for my $port (@workerPorts)
         {
-            system("$bindir/pg_ctl",
-                   ('stop', '-w', '-D', "tmp_check/worker.$port/data")) == 0
+            system(catfile("$bindir", "pg_ctl"),
+                   ('stop', '-w', '-D', catfile("tmp_check", "worker.$port", "data"))) == 0
                 or warn "Could not shutdown worker server";
         }
 
         if ($followercluster)
         {
-            system("$bindir/pg_ctl",
-                   ('stop', '-w', '-D', 'tmp_check/master-follower/data')) == 0
+            system(catfile("$bindir", "pg_ctl"),
+                   ('stop', '-w', '-D', catfile('tmp_check', 'master-follower', 'data'))) == 0
                 or warn "Could not shutdown worker server";
 
             for my $port (@followerWorkerPorts)
             {
-                system("$bindir/pg_ctl",
-                       ('stop', '-w', '-D', "tmp_check/follower.$port/data")) == 0
+                system(catfile("$bindir", "pg_ctl"),
+                       ('stop', '-w', '-D', catfile("tmp_check", "follower.$port", "data"))) == 0
                     or warn "Could not shutdown worker server";
             }
         }
+	if ($mitmPid != 0)
+	{
+	  # '-' means signal the process group, 2 is SIGINT
+          kill(-2, $mitmPid) or warn "could not interrupt mitmdump";
+	}
         $serversAreShutdown = "TRUE";
     }
 }
 
+if ($useMitmproxy)
+{
+  if (! -e $mitmFifoPath)
+  {
+    mkfifo($mitmFifoPath, 0777) or die "could not create fifo";
+  }
+
+  if (! -p $mitmFifoPath)
+  {
+    die "a file already exists at $mitmFifoPath, delete it before trying again";
+  }
+
+  my $childPid = fork();
+
+  die("Failed to fork\n")
+   unless (defined $childPid);
+
+  die("No child process\n")
+    if ($childPid < 0);
+
+  $mitmPid = $childPid;
+
+  if ($mitmPid eq 0) {
+    setpgrp(0,0); # we're about to spawn both a shell and a mitmdump, kill them as a group
+    exec("mitmdump --rawtcp -p 57640 --mode reverse:localhost:57638 -s mitmscripts/fluent.py --set fifo=$mitmFifoPath --set flow_detail=0 --set termlog_verbosity=warn >proxy.output 2>&1");
+    die 'could not start mitmdump';
+  }
+}
+
+$SIG{CHLD} = sub {
+  while ((my $waitpid = waitpid(-1, WNOHANG)) > 0) {}
+}; # If, for some reason, mitmproxy dies before we do
+
 # Set signals to shutdown servers
 $SIG{INT} = \&ShutdownServers;
 $SIG{QUIT} = \&ShutdownServers;
@@ -429,28 +605,29 @@ if ($valgrind)
     replace_postgres();
 }
 
+
 # Signal that servers should be shutdown
 $serversAreShutdown = "FALSE";
 
 # Start servers
-if(system("$bindir/pg_ctl",
+if(system(catfile("$bindir", "pg_ctl"),
        ('start', '-w',
         '-o', join(" ", @pgOptions)." -c port=$masterPort",
-       '-D', 'tmp_check/master/data', '-l', 'tmp_check/master/log/postmaster.log')) != 0)
+       '-D', catfile('tmp_check', 'master', 'data'), '-l', catfile('tmp_check', 'master', 'log', 'postmaster.log'))) != 0)
 {
-  system("tail", ("-n20", "tmp_check/master/log/postmaster.log"));
+  system("tail", ("-n20", catfile("tmp_check", "master", "log", "postmaster.log")));
   die "Could not start master server";
 }
 
 for my $port (@workerPorts)
 {
-    if(system("$bindir/pg_ctl",
+    if(system(catfile("$bindir", "pg_ctl"),
            ('start', '-w',
             '-o', join(" ", @pgOptions)." -c port=$port",
-            '-D', "tmp_check/worker.$port/data",
-            '-l', "tmp_check/worker.$port/log/postmaster.log")) != 0)
+            '-D', catfile("tmp_check", "worker.$port", "data"),
+            '-l', catfile("tmp_check", "worker.$port", "log", "postmaster.log"))) != 0)
     {
-      system("tail", ("-n20", "tmp_check/worker.$port/log/postmaster.log"));
+      system("tail", ("-n20", catfile("tmp_check", "worker.$port", "log", "postmaster.log")));
       die "Could not start worker server";
     }
 }
@@ -460,8 +637,8 @@ if ($followercluster)
 {
     # This test would run faster on PG10 if we could pass --no-sync here but that flag
     # isn't supported on PG 9.6. In a year when we drop support for PG9.6 add that flag!
-    system("$bindir/pg_basebackup",
-           ("-D", "tmp_check/master-follower/data", "--host=$host", "--port=$masterPort",
+    system(catfile("$bindir", "pg_basebackup"),
+           ("-D", catfile("tmp_check", "master-follower", "data"), "--host=$host", "--port=$masterPort",
             "--username=$user", "-R", "-X", "stream")) == 0
       or die 'could not take basebackup';
 
@@ -469,30 +646,30 @@ if ($followercluster)
     {
         my $workerPort = $workerPorts[$offset];
         my $followerPort = $followerWorkerPorts[$offset];
-        system("$bindir/pg_basebackup",
-               ("-D", "tmp_check/follower.$followerPort/data", "--host=$host", "--port=$workerPort",
+        system(catfile("$bindir", "pg_basebackup"),
+               ("-D", catfile("tmp_check", "follower.$followerPort", "data"), "--host=$host", "--port=$workerPort",
                 "--username=$user", "-R", "-X", "stream")) == 0
             or die "Could not take basebackup";
     }
 
-    if(system("$bindir/pg_ctl",
+    if(system(catfile("$bindir", "pg_ctl"),
            ('start', '-w',
             '-o', join(" ", @pgOptions)." -c port=$followerCoordPort",
-           '-D', 'tmp_check/master-follower/data', '-l', 'tmp_check/master-follower/log/postmaster.log')) != 0)
+           '-D', catfile('tmp_check', 'master-follower', 'data'), '-l', catfile('tmp_check', 'master-follower', 'log', 'postmaster.log'))) != 0)
     {
-      system("tail", ("-n20", "tmp_check/master-follower/log/postmaster.log"));
+      system("tail", ("-n20", catfile("tmp_check", "master-follower", "log", "postmaster.log")));
       die "Could not start master follower server";
     }
 
     for my $port (@followerWorkerPorts)
     {
-        if(system("$bindir/pg_ctl",
+        if(system(catfile("$bindir", "pg_ctl"),
                ('start', '-w',
                 '-o', join(" ", @pgOptions)." -c port=$port",
-                '-D', "tmp_check/follower.$port/data",
-                '-l', "tmp_check/follower.$port/log/postmaster.log")) != 0)
+                '-D', catfile("tmp_check", "follower.$port", "data"),
+                '-l', catfile("tmp_check", "follower.$port", "log", "postmaster.log"))) != 0)
         {
-          system("tail", ("-n20", "tmp_check/follower.$port/log/postmaster.log"));
+          system("tail", ("-n20", catfile("tmp_check", "follower.$port", "log", "postmaster.log")));
           die "Could not start follower server";
 	}
     }
@@ -504,55 +681,55 @@ if ($followercluster)
 ###
 for my $port (@workerPorts)
 {
-    system("psql", '-X',
-           ('-h', $host, '-p', $port, '-U', $user, "postgres",
+    system(catfile($bindir, "psql"),
+           ('-X', '-h', $host, '-p', $port, '-U', $user, "-d", "postgres",
             '-c', "CREATE DATABASE regression;")) == 0
         or die "Could not create regression database on worker";
 
     for my $extension (@extensions)
     {
-        system("psql", '-X',
-               ('-h', $host, '-p', $port, '-U', $user, "regression",
-                '-c', "CREATE EXTENSION IF NOT EXISTS \"$extension\";")) == 0
+        system(catfile($bindir, "psql"),
+               ('-X', '-h', $host, '-p', $port, '-U', $user, "-d", "regression",
+                '-c', "CREATE EXTENSION IF NOT EXISTS $extension;")) == 0
             or die "Could not create extension on worker";
     }
 
     foreach my $dataType (keys %dataTypes)
     {
-        system("psql", '-X',
-                ('-h', $host, '-p', $port, '-U', $user, "regression",
+        system(catfile($bindir, "psql"),
+                ('-X', '-h', $host, '-p', $port, '-U', $user, "-d", "regression",
                  '-c', "CREATE TYPE $dataType AS $dataTypes{$dataType};")) == 0
             or die "Could not create TYPE $dataType on worker";
     }
 
     foreach my $function (keys %functions)
     {
-        system("psql", '-X',
-                ('-h', $host, '-p', $port, '-U', $user, "regression",
+        system(catfile($bindir, "psql"),
+                ('-X', '-h', $host, '-p', $port, '-U', $user, "-d", "regression",
                  '-c', "CREATE FUNCTION $function RETURNS $functions{$function};")) == 0
             or die "Could not create FUNCTION $function on worker";
     }
 
     foreach my $operator (keys %operators)
     {
-        system("psql", '-X',
-                ('-h', $host, '-p', $port, '-U', $user, "regression",
+        system(catfile($bindir, "psql"),
+                ('-X', '-h', $host, '-p', $port, '-U', $user, "-d", "regression",
                  '-c', "CREATE OPERATOR $operator $operators{$operator};")) == 0
             or die "Could not create OPERATOR $operator on worker";
     }
 
     foreach my $fdw (keys %fdws)
     {
-        system("psql", '-X',
-                ('-h', $host, '-p', $port, '-U', $user, "regression",
+        system(catfile($bindir, "psql"),
+                ('-X', '-h', $host, '-p', $port, '-U', $user, "-d", "regression",
                  '-c', "CREATE FOREIGN DATA WRAPPER $fdw HANDLER $fdws{$fdw};")) == 0
             or die "Could not create foreign data wrapper $fdw on worker";
     }
 
     foreach my $fdwServer (keys %fdwServers)
     {
-        system("psql", '-X',
-                ('-h', $host, '-p', $port, '-U', $user, "regression",
+        system(catfile($bindir, "psql"),
+                ('-X', '-h', $host, '-p', $port, '-U', $user, "-d", "regression",
                  '-c', "CREATE SERVER $fdwServer FOREIGN DATA WRAPPER $fdwServers{$fdwServer};")) == 0
             or die "Could not create server $fdwServer on worker";
     }
@@ -563,7 +740,7 @@ my @arguments = (
     "--host", $host,
     '--port', $masterPort,
     '--user', $user,
-    '--bindir', "tmp_check/tmp-bin"
+    '--bindir', catfile("tmp_check", "tmp-bin")
 );
 
 # Add load extension parameters to the argument list
@@ -584,7 +761,7 @@ if ($vanillatest)
     $ENV{PGPORT} = $masterPort;
     $ENV{PGUSER} = $user;
 
-    system("make -C $postgresBuilddir/src/test/regress installcheck-parallel") == 0
+    system("make", ("-C", catfile("$postgresBuilddir", "src", "test", "regress"), "installcheck-parallel")) == 0
     or die "Could not run vanilla tests";
 }
 elsif ($isolationtester)
