@@ -11,7 +11,7 @@ CREATE TABLE uniq_cns_append_tables
 	partition_col integer UNIQUE,
 	other_col integer
 );
-SELECT master_create_distributed_table('uniq_cns_append_tables', 'partition_col', 'append');
+SELECT create_distributed_table('uniq_cns_append_tables', 'partition_col', 'append');
 
 CREATE TABLE excl_cns_append_tables
 (
@@ -19,7 +19,7 @@ CREATE TABLE excl_cns_append_tables
 	other_col integer,
         EXCLUDE (partition_col WITH =)
 );
-SELECT master_create_distributed_table('excl_cns_append_tables', 'partition_col', 'append');
+SELECT create_distributed_table('excl_cns_append_tables', 'partition_col', 'append');
 
 -- test that Citus cannot distribute unique constraints that do not include
 -- the partition column on hash-partitioned tables.
@@ -29,14 +29,14 @@ CREATE TABLE pk_on_non_part_col
 	partition_col integer,
 	other_col integer PRIMARY KEY
 );
-SELECT master_create_distributed_table('pk_on_non_part_col', 'partition_col', 'hash');
+SELECT create_distributed_table('pk_on_non_part_col', 'partition_col', 'hash');
 
 CREATE TABLE uq_on_non_part_col
 (
 	partition_col integer,
 	other_col integer UNIQUE
 );
-SELECT master_create_distributed_table('uq_on_non_part_col', 'partition_col', 'hash');
+SELECT create_distributed_table('uq_on_non_part_col', 'partition_col', 'hash');
 
 CREATE TABLE ex_on_non_part_col
 (
@@ -44,7 +44,7 @@ CREATE TABLE ex_on_non_part_col
 	other_col integer,
 	EXCLUDE (other_col WITH =)
 );
-SELECT master_create_distributed_table('ex_on_non_part_col', 'partition_col', 'hash');
+SELECT create_distributed_table('ex_on_non_part_col', 'partition_col', 'hash');
 
 -- now show that Citus can distribute unique and EXCLUDE constraints that 
 -- include the partition column for hash-partitioned tables.
@@ -57,14 +57,14 @@ CREATE TABLE pk_on_part_col
 	partition_col integer PRIMARY KEY,
 	other_col integer
 );
-SELECT master_create_distributed_table('pk_on_part_col', 'partition_col', 'hash');
+SELECT create_distributed_table('pk_on_part_col', 'partition_col', 'hash');
 
 CREATE TABLE uq_part_col
 (
 	partition_col integer UNIQUE,
 	other_col integer
 );
-SELECT master_create_distributed_table('uq_part_col', 'partition_col', 'hash');
+SELECT create_distributed_table('uq_part_col', 'partition_col', 'hash');
 
 CREATE TABLE uq_two_columns
 (
@@ -72,8 +72,7 @@ CREATE TABLE uq_two_columns
 	other_col integer,
 	UNIQUE (partition_col, other_col)
 );
-SELECT master_create_distributed_table('uq_two_columns', 'partition_col', 'hash');
-SELECT master_create_worker_shards('uq_two_columns', '4', '2');
+SELECT create_distributed_table('uq_two_columns', 'partition_col', 'hash');
 INSERT INTO uq_two_columns (partition_col, other_col) VALUES (1,1);
 INSERT INTO uq_two_columns (partition_col, other_col) VALUES (1,1);
 
@@ -83,8 +82,7 @@ CREATE TABLE ex_on_part_col
 	other_col integer,
 	EXCLUDE (partition_col WITH =)
 );
-SELECT master_create_distributed_table('ex_on_part_col', 'partition_col', 'hash');
-SELECT master_create_worker_shards('ex_on_part_col', '4', '2');
+SELECT create_distributed_table('ex_on_part_col', 'partition_col', 'hash');
 INSERT INTO ex_on_part_col (partition_col, other_col) VALUES (1,1);
 INSERT INTO ex_on_part_col (partition_col, other_col) VALUES (1,2);
 
@@ -94,8 +92,7 @@ CREATE TABLE ex_on_two_columns
 	other_col integer,
 	EXCLUDE (partition_col WITH =, other_col WITH =)
 );
-SELECT master_create_distributed_table('ex_on_two_columns', 'partition_col', 'hash');
-SELECT master_create_worker_shards('ex_on_two_columns', '4', '2');
+SELECT create_distributed_table('ex_on_two_columns', 'partition_col', 'hash');
 INSERT INTO ex_on_two_columns (partition_col, other_col) VALUES (1,1);
 INSERT INTO ex_on_two_columns (partition_col, other_col) VALUES (1,1);
 
@@ -105,8 +102,7 @@ CREATE TABLE ex_on_two_columns_prt
 	other_col integer,
 	EXCLUDE (partition_col WITH =, other_col WITH =) WHERE (other_col > 100)
 );
-SELECT master_create_distributed_table('ex_on_two_columns_prt', 'partition_col', 'hash');
-SELECT master_create_worker_shards('ex_on_two_columns_prt', '4', '2');
+SELECT create_distributed_table('ex_on_two_columns_prt', 'partition_col', 'hash');
 INSERT INTO ex_on_two_columns_prt (partition_col, other_col) VALUES (1,1);
 INSERT INTO ex_on_two_columns_prt (partition_col, other_col) VALUES (1,1);
 INSERT INTO ex_on_two_columns_prt (partition_col, other_col) VALUES (1,101);
@@ -118,7 +114,7 @@ CREATE TABLE ex_wrong_operator
 	other_col tsrange,
 	EXCLUDE USING gist (other_col WITH =, partition_col WITH &&)
 );
-SELECT master_create_distributed_table('ex_wrong_operator', 'partition_col', 'hash');
+SELECT create_distributed_table('ex_wrong_operator', 'partition_col', 'hash');
 
 CREATE TABLE ex_overlaps
 (
@@ -126,8 +122,7 @@ CREATE TABLE ex_overlaps
 	other_col tsrange,
 	EXCLUDE USING gist (other_col WITH &&, partition_col WITH =)
 );
-SELECT master_create_distributed_table('ex_overlaps', 'partition_col', 'hash');
-SELECT master_create_worker_shards('ex_overlaps', '4', '2');
+SELECT create_distributed_table('ex_overlaps', 'partition_col', 'hash');
 INSERT INTO ex_overlaps (partition_col, other_col) VALUES ('[2016-01-01 00:00:00, 2016-02-01 00:00:00]', '[2016-01-01 00:00:00, 2016-02-01 00:00:00]');
 INSERT INTO ex_overlaps (partition_col, other_col) VALUES ('[2016-01-01 00:00:00, 2016-02-01 00:00:00]', '[2016-01-15 00:00:00, 2016-02-01 00:00:00]');
 
@@ -142,14 +137,14 @@ CREATE TABLE pk_on_part_col_named
 	partition_col integer CONSTRAINT pk_on_part_col_named_pk PRIMARY KEY,
 	other_col integer
 );
-SELECT master_create_distributed_table('pk_on_part_col_named', 'partition_col', 'hash');
+SELECT create_distributed_table('pk_on_part_col_named', 'partition_col', 'hash');
 
 CREATE TABLE uq_part_col_named
 (
 	partition_col integer CONSTRAINT uq_part_col_named_uniq UNIQUE,
 	other_col integer
 );
-SELECT master_create_distributed_table('uq_part_col_named', 'partition_col', 'hash');
+SELECT create_distributed_table('uq_part_col_named', 'partition_col', 'hash');
 
 CREATE TABLE uq_two_columns_named
 (
@@ -157,8 +152,7 @@ CREATE TABLE uq_two_columns_named
 	other_col integer,
 	CONSTRAINT uq_two_columns_named_uniq UNIQUE (partition_col, other_col)
 );
-SELECT master_create_distributed_table('uq_two_columns_named', 'partition_col', 'hash');
-SELECT master_create_worker_shards('uq_two_columns_named', '4', '2');
+SELECT create_distributed_table('uq_two_columns_named', 'partition_col', 'hash');
 INSERT INTO uq_two_columns_named (partition_col, other_col) VALUES (1,1);
 INSERT INTO uq_two_columns_named (partition_col, other_col) VALUES (1,1);
 
@@ -168,8 +162,7 @@ CREATE TABLE ex_on_part_col_named
 	other_col integer,
 	CONSTRAINT ex_on_part_col_named_exclude EXCLUDE (partition_col WITH =)
 );
-SELECT master_create_distributed_table('ex_on_part_col_named', 'partition_col', 'hash');
-SELECT master_create_worker_shards('ex_on_part_col_named', '4', '2');
+SELECT create_distributed_table('ex_on_part_col_named', 'partition_col', 'hash');
 INSERT INTO ex_on_part_col_named (partition_col, other_col) VALUES (1,1);
 INSERT INTO ex_on_part_col_named (partition_col, other_col) VALUES (1,2);
 
@@ -179,8 +172,7 @@ CREATE TABLE ex_on_two_columns_named
 	other_col integer,
 	CONSTRAINT ex_on_two_columns_named_exclude EXCLUDE (partition_col WITH =, other_col WITH =)
 );
-SELECT master_create_distributed_table('ex_on_two_columns_named', 'partition_col', 'hash');
-SELECT master_create_worker_shards('ex_on_two_columns_named', '4', '2');
+SELECT create_distributed_table('ex_on_two_columns_named', 'partition_col', 'hash');
 INSERT INTO ex_on_two_columns_named (partition_col, other_col) VALUES (1,1);
 INSERT INTO ex_on_two_columns_named (partition_col, other_col) VALUES (1,1);
 
@@ -192,8 +184,7 @@ CREATE TABLE ex_multiple_excludes
 	CONSTRAINT ex_multiple_excludes_excl1 EXCLUDE (partition_col WITH =, other_col WITH =),
 	CONSTRAINT ex_multiple_excludes_excl2 EXCLUDE (partition_col WITH =, other_other_col WITH =)
 );
-SELECT master_create_distributed_table('ex_multiple_excludes', 'partition_col', 'hash');
-SELECT master_create_worker_shards('ex_multiple_excludes', '4', '2');
+SELECT create_distributed_table('ex_multiple_excludes', 'partition_col', 'hash');
 INSERT INTO ex_multiple_excludes (partition_col, other_col, other_other_col) VALUES (1,1,1);
 INSERT INTO ex_multiple_excludes (partition_col, other_col, other_other_col) VALUES (1,1,2);
 INSERT INTO ex_multiple_excludes (partition_col, other_col, other_other_col) VALUES (1,2,1);
@@ -204,7 +195,7 @@ CREATE TABLE ex_wrong_operator_named
 	other_col tsrange,
 	CONSTRAINT ex_wrong_operator_named_exclude EXCLUDE USING gist (other_col WITH =, partition_col WITH &&)
 );
-SELECT master_create_distributed_table('ex_wrong_operator_named', 'partition_col', 'hash');
+SELECT create_distributed_table('ex_wrong_operator_named', 'partition_col', 'hash');
 
 CREATE TABLE ex_overlaps_named
 (
@@ -212,8 +203,7 @@ CREATE TABLE ex_overlaps_named
 	other_col tsrange,
 	CONSTRAINT ex_overlaps_operator_named_exclude EXCLUDE USING gist (other_col WITH &&, partition_col WITH =)
 );
-SELECT master_create_distributed_table('ex_overlaps_named', 'partition_col', 'hash');
-SELECT master_create_worker_shards('ex_overlaps_named', '4', '2');
+SELECT create_distributed_table('ex_overlaps_named', 'partition_col', 'hash');
 INSERT INTO ex_overlaps_named (partition_col, other_col) VALUES ('[2016-01-01 00:00:00, 2016-02-01 00:00:00]', '[2016-01-01 00:00:00, 2016-02-01 00:00:00]');
 INSERT INTO ex_overlaps_named (partition_col, other_col) VALUES ('[2016-01-01 00:00:00, 2016-02-01 00:00:00]', '[2016-01-15 00:00:00, 2016-02-01 00:00:00]');
 
@@ -224,7 +214,7 @@ CREATE TABLE uq_range_tables
 	partition_col integer UNIQUE,
 	other_col integer
 );
-SELECT master_create_distributed_table('uq_range_tables', 'partition_col', 'range');
+SELECT create_distributed_table('uq_range_tables', 'partition_col', 'range');
 
 -- show that CHECK constraints are distributed.
 CREATE TABLE check_example
@@ -233,12 +223,10 @@ CREATE TABLE check_example
 	other_col integer CHECK (other_col >= 100),
 	other_other_col integer CHECK (abs(other_other_col) >= 100)
 );
-SELECT master_create_distributed_table('check_example', 'partition_col', 'hash');
-SELECT master_create_worker_shards('check_example', '2', '2');
-
+SELECT create_distributed_table('check_example', 'partition_col', 'hash');
 \c - - - :worker_1_port
-\d check_example_partition_col_key_365040
-SELECT "Constraint", "Definition" FROM table_checks WHERE relid='public.check_example_365040'::regclass;
+\d check_example_partition_col_key_365056
+SELECT "Constraint", "Definition" FROM table_checks WHERE relid='public.check_example_365056'::regclass;
 \c - - - :master_port
 
 -- drop unnecessary tables
